/* ckbd_drv.c
 *
 * Interface to Cherry Keyboards
 *
 */

#include <linux/version.h>
#include <linux/kernel.h>   
#include <linux/module.h>  


/* Deal with CONFIG_MODVERSIONS */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
//#include <linux/modversions.h>
#endif

/* For character devices */

/* The character device definitions are here */
#include <linux/fs.h>

/* A wrapper which does next to nothing at
 * at present, but may help for compatibility
 * with future versions of Linux */
//#include <linux/wrapper.h>


//#include <linux/kbd_ll.h>
#include <linux/sched.h>

/* Our own definitions */
#include "ckbd_drv.h"

/* In 2.2.3 /usr/include/linux/version.h includes a
 * macro for this, but 2.0.35 doesn't - so I add it
 * here if necessary. */
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>  /* for get_user and put_user */
#endif

#define SUCCESS 0



/* to output some debug messages */
//#define MY_DEBUG

/* to output some debug messages by the keyhook function */
#ifdef MY_DEBUG
#define MY_DEBUG_KBDHOOK
#endif

/* to output some debug messages by the keyhook function for MSR protocoll */

/* Defines */
#define XON	0x11
#define XOFF	0x13

#define HISTORY_LEN 6
#define STAT_LEN 8
#define MSR_LEN 512
/* Device Declarations ******************************** */

MODULE_AUTHOR("2004 Cherry GmbH");
MODULE_DESCRIPTION("Interface to Cherry Keyboards");
MODULE_SUPPORTED_DEVICE("Cherry Keyboards");
MODULE_LICENSE("GPL");

/* for timeout timer */
struct timer_list jiq_timer;

/* in kernel */
extern int (*kbdhook)(unsigned char);

// send_data function doesnt exists anymore
//extern int send_data(unsigned char data);
extern struct ps2dev *kbddevice;
extern int ps2_sendbyte(struct ps2dev *ps2dev,unsigned char byte, int timeout);
extern void kbd_keycode(unsigned int keycode, int down, int hw_raw);

/* Prototypes ***************************************** */
static int my_kbdhook(unsigned char);
static int send2kbd(struct ps2dev *ps2dev, unsigned char c);

/* The name for our device, as it will appear in
 * /proc/devices */
#define DEVICE_NAME "ckbd_dev"


/* The maximum length of the message for the device */
#define BUF_LEN 128

/* Is the device open right now? Used to prevent
 * concurent access into the same device */
static int Device_Open = 0;
static int Device_Minor = 0;
static int Device_Major = 0;

static spinlock_t       ckbd_u_lock;
static int ckbd_u_count = 0;
static unsigned int ckbd_u_owner = 0;

static int ckbd_u_stat = 0;
static int ckbd_u_keyboard = 0;
static int ckbd_u_msr1 = 0;
static int ckbd_u_msr2 = 0;
static int ckbd_u_msr3 = 0;

/* The message the device will give when asked */
static char Message[BUF_LEN];

/* scancode history for protokoll separation */
static unsigned char scancode_history[HISTORY_LEN];
static unsigned char ignore_data;

/* How far did the process reading the message get?
 * Useful if the message is larger than the size of the
 * buffer we get to fill in device_read. */
static char *Message_Ptr;
static size_t Message_Len = 0;

/* Keylock data area */
/* keylockcodes
  e0 12 e0 92 - stellung 1
  e0 05 e0 85 - stellung 2
  e0 02 e0 82 - stellung 3
  e0 03 e0 83 - stellung 4
  e0 04 e0 84 - stellung 5
*/
static unsigned char keylock1[] = { 0xe0,0x12,0xe0,0x92};/* Keylock position 1*/
static unsigned char keylock2[] = { 0xe0,0x05,0xe0,0x85};/* Keylock position 2*/
static unsigned char keylock3[] = { 0xe0,0x02,0xe0,0x82};/* Keylock position 3*/
static unsigned char keylock4[] = { 0xe0,0x03,0xe0,0x83};/* Keylock position 4*/
static unsigned char keylock5[] = { 0xe0,0x04,0xe0,0x84};/* Keylock position 5*/

/* MSRx data area */
static unsigned char msr_header[] = { 0xe0,0x1d,0x36,0x1d,0x2a};/* header MSR1 start*/
static unsigned char msr_footer[] = { 0xaa,0x9d,0xb6,0xe0,0x9d};/* header MSR1 start*/
static unsigned char msr1_start[] = { 0x02,0x82};/* header MSR1 start*/
static unsigned char msr2_start[] = { 0x03,0x83};/* header MSR2 start*/
static unsigned char msr3_start[] = { 0x04,0x84};/* header MSR3 start*/
static unsigned char msr1_stop[]  = { 0x09,0x89};/* header MSR1 stop*/
static unsigned char msr2_stop[]  = { 0x0a,0x8a};/* header MSR2 stop*/
static unsigned char msr3_stop[]  = { 0x0b,0x8b};/* header MSR3 stop*/


static unsigned char msr_zustand;
static char MSR_track1[MSR_LEN], MSR_track2[MSR_LEN], MSR_track3[MSR_LEN];
static int  MSR_len1 = 0, MSR_len2 = 0, MSR_len3 = 0;
static char *pTrack1Data, *pTrack2Data, *pTrack3Data;


/* STAT data area */
static int  capture_keyboard = 0;
static unsigned char breakkey[] = { 0xe1,0x1d,0x45,0xe1,0x9d,0xc5};/* Keylock position 1*/
static struct {
  unsigned char keylock; /* keylock status 1-3 */
  unsigned char scancodenr; /* number of scancodes in buffer */
  short int msr1nr; /* number of bytes in msr1 */
  short int msr2nr; /* number of bytes in msr2 */
  short int msr3nr; /* number of bytes in msr3 */
} StatData;
 
static int STAT_len = 0;
static unsigned  char *pStatData;
#define KEYLOCK_STATUS  StatData.keylock
#define KEYBOARD_STATUS  StatData.scancodenr

/* Keyboard data area */
static unsigned char KeyboardData[BUF_LEN];
static unsigned char *pKeyboardData;
static unsigned char KeyboardLen;
static unsigned char scanstat;

#define KEYCONF_LEN 385
/* block internal handling of scancodes
offset 0-127 standard scancodes
offset 128 - 255 e0 scancodes
offset 256 - 383 e1 scancodes */
static unsigned char keyconf[KEYCONF_LEN];

/* This function is called whenever a process attempts
 * to open the device file */
static int device_open(struct inode *inode,
                       struct file *file)
{
#ifdef MY_DEBUG
  printk("device_open(%p)\n", file);
#endif

  /* We don't want to talk to two processes at the
   * same time */
  //if (Device_Open)
  //  return -EBUSY;
  /* We restrict the access to a single user */
  spin_lock(&ckbd_u_lock);
  if(ckbd_u_count &&
     (ckbd_u_owner != current->uid) &&  /* allow user */
     (ckbd_u_owner != current->euid) && /* allow whoever did su */
     !capable(CAP_DAC_OVERRIDE)) {      /* still allow root */
    spin_unlock(&ckbd_u_lock);
    return -EBUSY;      /* -EPERM would confuse the user */
  }

  if(ckbd_u_count == 0) {
    ckbd_u_owner = current->uid;        /* grab it */
  }

  ckbd_u_count++;
  spin_unlock(&ckbd_u_lock);


  /* If this was a process, we would have had to be
   * more careful here, because one process might have
   * checked Device_Open right before the other one
   * tried to increment it. However, we're in the
   * kernel, so we're protected against context switches.
   *
   * This is NOT the right attitude to take, because we
   * might be running on an SMP box, but we'll deal with
   * SMP in a later chapter.
   */

  Device_Open++;
  Device_Minor = MINOR(inode->i_rdev);
#ifdef MY_DEBUG
  printk ("device_open minor = %d\n",Device_Minor);
#endif
  switch(Device_Minor) {
    case STAT_MINOR_NUM:
      ckbd_u_stat++;
      break;
    case KEYBOARD_MINOR_NUM:
      ckbd_u_keyboard++;
      pKeyboardData = &KeyboardData[0];
      KeyboardLen = 0;
      break;
    case MSR1_MINOR_NUM:
      ckbd_u_msr1++;
      break;
    case MSR2_MINOR_NUM:
      ckbd_u_msr2++;
      break;
    case MSR3_MINOR_NUM:
      ckbd_u_msr3++;
      break;
  }
  /* Initialize the message */
  Message_Ptr = Message;
  Message_Len = 0;

  //MOD_INC_USE_COUNT;

  return SUCCESS;
}


/* This function is called when a process closes the
 * device file. It doesn't have a return value because
 * it cannot fail. Regardless of what else happens, you
 * should always be able to close a device (in 2.0, a 2.2
 * device file could be impossible to close). */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int device_release(struct inode *inode,
                          struct file *file)
#else
static void device_release(struct inode *inode,
                           struct file *file)
#endif
{
#ifdef MY_DEBUG
  printk ("device_release(%p,%p)\n", inode, file);
#endif
  Device_Minor = MINOR(inode->i_rdev);
#ifdef MY_DEBUG
  printk ("device_release minor = %d\n",Device_Minor);
#endif
  switch(Device_Minor) {
    case STAT_MINOR_NUM:
      ckbd_u_stat--;
      break;
    case KEYBOARD_MINOR_NUM:
      ckbd_u_keyboard--;
      break;
    case MSR1_MINOR_NUM:
      ckbd_u_msr1--;
      break;
    case MSR2_MINOR_NUM:
      ckbd_u_msr2--;
      break;
    case MSR3_MINOR_NUM:
      ckbd_u_msr3--;
      break;
  }

  /* We're now ready for our next caller */
  Device_Open --;
  ckbd_u_count--;

  //MOD_DEC_USE_COUNT;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  return 0;
#endif
}



/* This function is called whenever a process which
 * has already opened the device file attempts to
 * read from it. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t device_read(
    struct file *file,
    char *buffer,   /* The buffer to fill with the data */
    size_t length,  /* The length of the buffer */
    loff_t *offset) /* offset to the file */
{
  int bytes_read = 0;
  Device_Minor = MINOR(file->f_dentry->d_inode->i_rdev);
/*
#ifdef MY_DEBUG
  printk("device_read(%d,%p,%p,%d)\n", Device_Minor, file, buffer, length);
#endif
*/
#else
static int device_read(
    struct inode *inode,
    struct file *file,
    char *buffer,   /* The buffer to fill with the data */
    int length)     /* The length of the buffer
                     * (mustn't write beyond that!) */
{
  int bytes_read = 0;
  Device_Minor = MINOR(inode->i_rdev);

#ifdef MY_DEBUG
  printk("device_read(%p,%p,%p,%d)\n", inode, file, buffer, length);
  printk("Device_Minor = %d\n",Device_Minor);
#endif

#endif

  /* Number of bytes actually written to the buffer */
  if(length <= 0)
    return -EINVAL;

  if (Device_Minor == STAT_MINOR_NUM) {   // STAT
#ifdef MY_DEBUG
    //printk("Read STAT data %d\n",STAT_len);
#endif
      if (length > BUF_LEN) length = BUF_LEN;
      pStatData = &StatData.keylock;
      Message_Ptr = pStatData;      //StatData
      STAT_len = STAT_LEN;
  }
  else if (Device_Minor == KEYBOARD_MINOR_NUM) {   // KEYBOARD
#ifdef MY_DEBUG
    printk("Read KEYBOARD data %d\n",length);
#endif
    if (length > BUF_LEN) length = BUF_LEN;
      pKeyboardData = &KeyboardData[0];
      Message_Ptr = pKeyboardData;
  }
  else if (Device_Minor == MSR1_MINOR_NUM) {   // MSR1
#ifdef MY_DEBUG
    printk("Read MSR track1 data %d\n",MSR_len1);
#endif
    if (length > MSR_LEN) length = MSR_LEN;
    if(pTrack1Data)
      Message_Ptr = pTrack1Data;      //MSR_track1
    else
      return 0;
  }
  else if (Device_Minor == MSR2_MINOR_NUM) {      // MSR2
#ifdef MY_DEBUG
    printk("Read MSR track2 data %d\n",MSR_len2);
#endif
    if (length > MSR_LEN) length = MSR_LEN;
    if(pTrack2Data)
      Message_Ptr = pTrack2Data;      //MSR_track2
    else
      return 0;
  }
  else if (Device_Minor == MSR3_MINOR_NUM) {     // MSR3
#ifdef MY_DEBUG
    printk("Read MSR track3 data %d\n",MSR_len3);
#endif
    if (length > MSR_LEN) length = MSR_LEN;
    if(pTrack3Data)
      Message_Ptr = pTrack3Data;      //MSR_track3;
    else
      return 0;
  }
  else if (Device_Minor == KEYLOCK_MINOR_NUM) {   // Keylock
#ifdef MY_DEBUG
    printk("Read Keylock Position %x\n",KEYLOCK_STATUS);
#endif
    Message[0] = KEYLOCK_STATUS;
    Message_Ptr = Message;
  }

  /* If we're at the end of the message, return 0
   * (which signifies end of file) */
  if (Message_Ptr == 0)
    return 0;

  /* Actually put the data into the buffer */
  while (length)  {

    /* Because the buffer is in the user data segment,
     * not the kernel data segment, assignment wouldn't
     * work. Instead, we have to use put_user which
     * copies data from the kernel data segment to the
     * user data segment. */
    put_user(*(Message_Ptr++), buffer++);
    length --;
    bytes_read ++;
  }

#ifdef MY_DEBUG
  printk ("Read %d bytes, %d left\n", bytes_read, length);
#endif

  /* ReInitialize the message */

  if (Device_Minor == MSR1_MINOR_NUM) {        // MSR1
    if(*Message_Ptr == 0) {
      pTrack1Data = 0;
      MSR_len1 = 0;
      MSR_track1[0] = 0;
    }
    else {
      pTrack1Data = Message_Ptr;
    }
    StatData.msr1nr = 0;
  }
  else if (Device_Minor == MSR2_MINOR_NUM) {        // MSR2
    if(*Message_Ptr == 0) {
      pTrack2Data = 0;
      MSR_len2 = 0;
      MSR_track2[0] = 0;
    }
    else {
      pTrack2Data = Message_Ptr;
    }
    StatData.msr2nr = 0;
  }
  else if (Device_Minor == MSR3_MINOR_NUM) {        // MSR3
    if(*Message_Ptr == 0) {
      pTrack3Data = 0;
      MSR_len3 = 0;
      MSR_track3[0] = 0;
    }
    else {
      pTrack3Data = Message_Ptr;
    }
    StatData.msr3nr = 0;
  }
  else if (Device_Minor == KEYBOARD_MINOR_NUM) {        // Keyboard
    pKeyboardData = &KeyboardData[0];
    KeyboardLen = 0;
    KEYBOARD_STATUS = 0;
  }

  /* Read functions are supposed to return the number
   * of bytes actually inserted into the buffer */
  return bytes_read;
}


/* This function is called when somebody tries to
 * write into our device file. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t device_write(struct file *file,
                            const char *buffer,
                            size_t length,
                            loff_t *offset)
#else
static int device_write(struct inode *inode,
                        struct file *file,
                        const char *buffer,
                        int length)
#endif
{
  unsigned int i=0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  Device_Minor = MINOR(file->f_dentry->d_inode->i_rdev);
#else
  Device_Minor = MINOR(inode->i_rdev);
#endif

#ifdef MY_DEBUG
  //printk ("device_write(%d,%p,%s,%d)\n", Device_Minor, file, buffer, length);
#endif

  /** Write to STAT
  *	send buffer to keyboard firmware
  */
  if(Device_Minor == STAT_MINOR_NUM) {   //STAT
      for(i=0; i<length && i<BUF_LEN; i++) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
        get_user(Message[i], buffer+i);
#else
        Message[i] = get_user(buffer+i);
#endif
	//printk(KERN_INFO "CherryPS2: device_write(): send2kbd(%x)", Message[i]);
	send2kbd(kbddevice,Message[i]);
      }
  }

  /** Write to KEYLOCK
  *	send buffer to system
  */
  else if(Device_Minor == KEYLOCK_MINOR_NUM) {   //KEYLOCK
    for(i=0; i<length && i<BUF_LEN; i++) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
      get_user(Message[i], buffer+i);
#else
      Message[i] = get_user(buffer+i);
#endif
	//printk(KERN_INFO "CherryPS2: device_write(): Write to Keylock: handle_scancode(%x)\n", Message[i]);
	kbd_keycode(Message[i],!(Message[i] & 0x80),0);	
    }
  }

  /** Write to POSKeyboard
  *	send buffer to POSKeyboard
  *	  used to enable/disable capture POSKeyboard data via XON/XOFF
  */
  else if(Device_Minor == KEYBOARD_MINOR_NUM) {   //POSKeyboard
    for(i=0; i<length && i<KEYCONF_LEN; i++) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
      get_user(keyconf[i], buffer+i);
#else
      keyconf[i] = get_user(buffer+i);
#endif
#ifdef MY_DEBUG
       printk("Key %x : Konfig %x\n",i,keyconf[i]);
#endif
    }
  }

  /** Write to other devices
  *
  */
  else {
    printk("device_write not available for this device %d\n",Device_Minor);
    i = 0;
  }

  Message_Ptr = Message;

  /* Again, return the number of input characters used */
  return i;
}


/* Functions ******************************** */
static int send2kbd(struct ps2dev *ps2dev, unsigned char b)
{
// in send_data timeout was set to KBD_TIMEOUT
// value of KBD_TIMEOUT is 0x100000 so timeout is set to 0x100000
	int iRes = 0;
	//printk(KERN_INFO "CherryPS2: send2kbd(): send_data(%x)\n", b);
	//printk(KERN_INFO "CherryPS2: ps2dev = %x \n", (unsigned int)ps2dev);
	if (ps2dev != NULL)
	{
		iRes = ps2_sendbyte(ps2dev,b,0x100000);
		if (!iRes)
		{
			//printk(KERN_INFO "CherryPS2: ERROR send2kbd(%x) iRes = %d \n", b, iRes);
			ps2_sendbyte(ps2dev,0x0F4,0x100000);
			return 1;
		}
		//printk(KERN_INFO "CherryPS2: SUCCESS send2kbd(%x)\n", b);
	}
	else
	{
		printk(KERN_INFO "CherryPS2: ps2dev is NULL! \n");
		return 1;
	}
  return 0;
}

/* Timer routine for timeout function */
void jiq_timedout(unsigned long ptr)
{
  // add commands for timeout handling
#ifdef MY_DEBUG
  printk("Timeout\n");
#endif
  scanstat = 0;
  msr_zustand = 0;
  ignore_data = 0;
}

/* My Keyboard Hook ****************************** */
static int my_kbdhook(unsigned char scancode)
{
unsigned int i;

#ifdef MY_DEBUG
  printk("Scancode %x\n",scancode);
#endif


  if (ignore_data)
  {
    if (ignore_data == 1)
    {
      KEYLOCK_STATUS = (scancode & 0x0f);
#ifdef MY_DEBUG_KBDHOOK
      printk("initial keylock status : %x\n", KEYLOCK_STATUS);
#endif
    }
    ignore_data--;
    return 1;
  }
  
   // retrigger timeout counter
#ifdef MY_DEBUG
  mod_timer(&jiq_timer, jiffies + 5 * HZ);	//retrigger timeout counter 5secs
#else
  mod_timer(&jiq_timer, jiffies + HZ/2);	//retrigger timeout counter 0.5secs
#endif

  for (i=0;i<HISTORY_LEN-1;i++)
  {
    scancode_history[i] = scancode_history[i+1];
  }
  scancode_history[HISTORY_LEN-1] = scancode;

  if (memcmp(&keylock1[0],&scancode_history[HISTORY_LEN - sizeof(keylock1)],sizeof(keylock1)) == 0)
  {
    KEYLOCK_STATUS = 0x01;
  }
  if (memcmp(&keylock2[0],&scancode_history[HISTORY_LEN - sizeof(keylock2)],sizeof(keylock2)) == 0)
  {
    KEYLOCK_STATUS = 0x02;
  }
  if (memcmp(&keylock3[0],&scancode_history[HISTORY_LEN - sizeof(keylock3)],sizeof(keylock3)) == 0)
  {
    KEYLOCK_STATUS = 0x03;
  }
  if (memcmp(&keylock4[0],&scancode_history[HISTORY_LEN - sizeof(keylock4)],sizeof(keylock4)) == 0)
  {
    KEYLOCK_STATUS = 0x04;
  }
  if (memcmp(&keylock5[0],&scancode_history[HISTORY_LEN - sizeof(keylock5)],sizeof(keylock5)) == 0)
  {
    KEYLOCK_STATUS = 0x05;
  }

  switch (msr_zustand)
  {
    case 0:
      if (memcmp(&msr_header[0],&scancode_history[HISTORY_LEN - sizeof(msr_header)],sizeof(msr_header)) == 0)
      {  /* ein msrheader empfangen*/
        msr_zustand = 1;
      }
      break;
      
    case 1: /* 1.byte msr 1 2 oder 3 */
      msr_zustand = 2;
      return 1;

    case 2: /* 2.byte msr 1 2 oder 3 */
      if (memcmp(&msr1_start[0],&scancode_history[HISTORY_LEN - sizeof(msr1_start)],sizeof(msr1_start)) == 0)
      {
        msr_zustand = 3;
      }
      if (memcmp(&msr2_start[0],&scancode_history[HISTORY_LEN - sizeof(msr2_start)],sizeof(msr2_start)) == 0)
      {
        msr_zustand = 4;
      }
      if (memcmp(&msr3_start[0],&scancode_history[HISTORY_LEN - sizeof(msr3_start)],sizeof(msr3_start)) == 0)
      {
        msr_zustand = 5;
      }
      return 1;
       
    case 3:/* waitfor footer*/
      if (memcmp(&msr_footer[0],&scancode_history[HISTORY_LEN - sizeof(msr_footer)],sizeof(msr_footer)) == 0)
      {  /* ein msrfooter empfangen*/
        msr_zustand = 6;
        StatData.msr1nr = -1;
#ifdef MY_DEBUG
        printk("Header MSR1 start found\n");
#endif
      }
      MSR_len1 = 0;
      return 0;
      break;

    case 4:/* waitfor footer*/
      if (memcmp(&msr_footer[0],&scancode_history[HISTORY_LEN - sizeof(msr_footer)],sizeof(msr_footer)) == 0)
      {  /* ein msrfooter empfangen*/
        msr_zustand = 7;
        StatData.msr2nr = -1;
  #ifdef MY_DEBUG
        printk("Header MSR2 start found\n");
#endif
      }
      MSR_len2 = 0;
      return 0;
      break;

    case 5:/* waitfor footer*/
      if (memcmp(&msr_footer[0],&scancode_history[HISTORY_LEN - sizeof(msr_footer)],sizeof(msr_footer)) == 0)
      {  /* ein msrfooter empfangen*/
        msr_zustand = 8;
        StatData.msr3nr = -1;
#ifdef MY_DEBUG
        printk("Header MSR3 start found\n");
#endif
      }
      MSR_len3 = 0;
      return 0;
      break;

    case 6:/* write msr1 data */
      if (memcmp(&msr_header[0],&scancode_history[HISTORY_LEN - sizeof(msr_header)],sizeof(msr_header)) == 0)
      {  /* ein msrheader empfangen*/
        pTrack1Data = &MSR_track1[0];
        MSR_len1 -= sizeof(msr_header)-1;
#ifdef MY_DEBUG
        printk("%x MSR1 Bytes empfangen\n",MSR_len1);
#endif
        StatData.msr1nr = MSR_len1;
        msr_zustand = 9;
        return 1;
      }
      MSR_track1[MSR_len1] = scancode;
      MSR_len1++;
      return 1;

    case 7:/* write msr2 data */
      if (memcmp(&msr_header[0],&scancode_history[HISTORY_LEN - sizeof(msr_header)],sizeof(msr_header)) == 0)
      {  /* ein msrheader empfangen*/
        pTrack2Data = &MSR_track2[0];
        MSR_len2 -= sizeof(msr_header)-1;
        StatData.msr2nr = MSR_len2;
        msr_zustand = 9;
        return 1;
       }
      MSR_track2[MSR_len2] = scancode;
      MSR_len2++;
      return 1;

    case 8:/* write msr3 data */
      if (memcmp(&msr_header[0],&scancode_history[HISTORY_LEN - sizeof(msr_header)],sizeof(msr_header)) == 0)
      {  /* ein msrheader empfangen*/
        pTrack3Data = &MSR_track3[0];
        MSR_len3 -= sizeof(msr_header)-1;
        StatData.msr3nr = MSR_len3;
        msr_zustand = 9;
        return 1;
      }
      MSR_track3[MSR_len3] = scancode;
      MSR_len3++;
      return 1;

    case 9:/* waitfor footer*/
      if (memcmp(&msr_footer[0],&scancode_history[HISTORY_LEN - sizeof(msr_footer)],sizeof(msr_footer)) == 0)
      {  /* ein msrfooter empfangen*/
        msr_zustand = 0;
      }
      return 1;

    default: msr_zustand = 0;
  }
   
  if (ckbd_u_keyboard)
  { /* keyboard open, use keykonfig */
    switch (scanstat)
    {
      case 0:
        if (scancode == 0xe0)
        {
          scanstat = 1;
#ifdef MY_DEBUG
          printk("Code e0 filtered\n");
#endif
          return 1;
        }
        if (scancode == 0xe1)
        {
          scanstat = 2;
#ifdef MY_DEBUG
          printk("Code e1 filtered\n");
#endif
          return 1;
        }
        if (keyconf[(scancode & 0x7f)])
        { /* suppress scancode for OS and give it to JPOS */
#ifdef MY_DEBUG
          printk("Key filtered %x\n",scancode);
#endif
          if (KeyboardLen < BUF_LEN)
          {
            KeyboardData[KeyboardLen++] = scancode;
          }
          KEYBOARD_STATUS = KeyboardLen;
          return 1;
        }
        break;

      case 1:/* delay e0 scancode */
        if (scancode == 0x2a) /*possible 4 byte scancode */
        {
          scanstat = 3;
          return 1;
        }
        
        scanstat = 0;
        if (keyconf[((scancode & 0x7f)+128)])
        { /* suppress scancode for OS and give it to JPOS */
#ifdef MY_DEBUG
          printk("Key filtered e0 %x\n",scancode);
#endif
          if (KeyboardLen < (BUF_LEN - 1))
          {
            KeyboardData[KeyboardLen++] = 0xe0;
            KeyboardData[KeyboardLen++] = scancode;
          }
          KEYBOARD_STATUS = KeyboardLen;
          return 1;
        }
        else {
		//printk(KERN_INFO "CherryPS2: my_kbdhook(): handle_scancode(0xe0)\n");
		kbd_keycode(0xe0,!(0xe0 & 0x80),0);

        }
        break;

      case 2:/* delay e1 scancode */
        if (scancode == 0x1d) /*6 byte scancode for break key*/
        {
          scanstat = 9;
          return 1;
        }
        scanstat = 0;
        if (keyconf[((scancode & 0x7f)+256)])
        { /* suppress scancode for OS and give it to JPOS */
#ifdef MY_DEBUG
          printk("Key filtered e1 %x\n",scancode);
#endif
          if (KeyboardLen < (BUF_LEN - 1))
          {
            KeyboardData[KeyboardLen++] = 0xe1;
            KeyboardData[KeyboardLen++] = scancode;
          }
          KEYBOARD_STATUS = KeyboardLen;
          return 1;
        }
        else {
		//printk(KERN_INFO "CherryPS2: my_kbdhook(): handle_scancode(0xe1)\n");
		kbd_keycode(0xe1,!(0xe1 & 0x80),0);
        }
        break;

      case 3:/* 4 byte scancode */
        scanstat = 4;
        return 1;
        
      case 4:/* 4 byte scancode */
        if (keyconf[((scancode & 0x7f)+128)])
        { /* suppress scancode for OS and give it to JPOS */
#ifdef MY_DEBUG
          printk("Key filtered e0 2a e0 %x\n",scancode);
#endif
          if (KeyboardLen < (BUF_LEN - 1))
          {
            KeyboardData[KeyboardLen++] = 0xe0;
            KeyboardData[KeyboardLen++] = scancode;
          }
          KEYBOARD_STATUS = KeyboardLen;
          scanstat = 5;
          return 1;
        }
	kbd_keycode(0xe0,!(0xe0 & 0x80),0);
	kbd_keycode(0x2a,!(0x2a & 0x80),0);
	kbd_keycode(0xe0,!(0xe0 & 0x80),0);
        scanstat = 7;
        break;

      case 5:/* wait for 4 byte scancode end e0 - no output*/
        scanstat = 6;
        return 1;

      case 6:/* wait for 4 byte scancode end aa - no output*/
        scanstat = 0;
        return 1;
                
      case 7:/* wait for 4 byte scancode end e0 - output*/
        scanstat = 8;
        break;

      case 8:/* wait for 4 byte scancode end aa - output*/
        scanstat = 0;
        break;

      case 9:/* wait for completed break key scancode*/
        if (memcmp(&breakkey[0],&scancode_history[HISTORY_LEN - sizeof(breakkey)],sizeof(breakkey)) == 0)
        {  /* breakkey complete*/
          if (keyconf[384])
          { 
            if (KeyboardLen < (BUF_LEN - sizeof(breakkey)))
            {
              for (i=0;i<sizeof(breakkey);i++)
              {
                KeyboardData[KeyboardLen++] = breakkey[i];
              }
            }
            KEYBOARD_STATUS = KeyboardLen;
          }
          else
          {
            for (i=0;i<sizeof(breakkey);i++)
            {
		//printk(KERN_INFO "CherryPS2: my_kbdhook(): handle_scancode(%x)\n", breakkey[i]);
		kbd_keycode(breakkey [i],!(breakkey[i] & 0x80),0);
            }
          }
          scanstat = 0;
        }
        return 1;

      default:
        scanstat = 0;
    }
  }
  return 0;
}


/* Module Declarations *************************** */


/* This structure will hold the functions to be called
 * when a process does something to the device we
 * created. Since a pointer to this structure is kept in
 * the devices table, it can't be local to
 * init_module. NULL is for unimplemented functions. */
struct file_operations Fops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
  owner:  THIS_MODULE,
#endif
  read:   device_read,
  write:  device_write,
  open:   device_open,
  release: device_release
};


/* Initialize the module - Register the character device */
int init_module()
{
  int ret_val;
  int i;

  /* Register the character device (atleast try) */
  ret_val = register_chrdev(CKBD_MAJOR_NUM,
                            DEVICE_NAME,
                            &Fops);

  /* Negative values signify an error */
  if (ret_val < 0) {
    printk ("%s failed with %d\n",
            "Sorry, registering the character device ",
            ret_val);
    return ret_val;
  }
  Device_Major = ret_val;

//  printk ("%s The major device number is %d.\n",
//          "Registeration is a success", Device_Major);
//  printk ("If you want to talk to the device driver,\n");
//  printk ("you'll have to create device files. \n");
//  printk ("We suggest you use:\n");
//  printk ("  mknod /dev/ckbd_stat c %d 0\n", Device_Major);
//  printk ("  mknod /dev/ckbd_keyboard c %d 1\n", Device_Major);
//  printk ("  mknod /dev/ckbd_msr1 c %d 2\n", Device_Major);
//  printk ("  mknod /dev/ckbd_msr2 c %d 3\n", Device_Major);
//  printk ("  mknod /dev/ckbd_msr3 c %d 4\n", Device_Major);
//  printk ("  mknod /dev/ckbd_keylock c %d 5\n", Device_Major);
//  printk ("The device file names are important, because\n");
//  printk ("the Cherry tools assume that are the\n");
//  printk ("files you'll use.\n");

  spin_lock_init(&ckbd_u_lock);   //init spin_lock

  kbdhook = my_kbdhook;
  printk ("Cherry PS/2 Module is active now! (%p)\n",kbdhook);

  /* init Keylock data */
  KEYLOCK_STATUS = 0;

  /* init MSR data */
  MSR_len1 = 0;
  MSR_len2 = 0;
  MSR_len3 = 0;
  pTrack1Data = 0;
  pTrack2Data = 0;
  pTrack3Data = 0;

  /* init STAT data */
  StatData.keylock = 0;
  StatData.scancodenr = 0;
  StatData.msr1nr = 0;
  StatData.msr2nr = 0;
  StatData.msr3nr = 0;
  STAT_len = 0;
  pStatData = 0;
  capture_keyboard = 0;

  /* init Keyboard data */
  KeyboardData[0] = 0;
  KEYBOARD_STATUS = 0;
  pKeyboardData = 0;

  // init timeout counter
  init_timer(&jiq_timer);
  jiq_timer.function = jiq_timedout;
  jiq_timer.data = 0;

  jiq_timer.expires = jiffies + 1 * HZ;	// 1 sec
  add_timer(&jiq_timer);
  del_timer_sync(&jiq_timer);

  for (i=0;i<KEYCONF_LEN;i++)
  {
    keyconf[i]=0;
  }
  
/* zum test F5 bis F8 hart auf JPOS setzen */
//    keyconf[0x3f]=1; /* F5 */
//    keyconf[0x40]=1; /* F6 */
//    keyconf[0x41]=1; /* F7 */
//    keyconf[0x42]=1; /* F8 */
/*******************************************/

  scanstat = 0;
  msr_zustand = 0;
  
  ignore_data = 16;

  //printk(KERN_INFO "TRYING TO SEND 0xc0 TO %x\n", (unsigned int)kbddevice);
  if (kbddevice != NULL)  
  	send2kbd(kbddevice,0xc0);


  mod_timer(&jiq_timer, jiffies + HZ/2);	//retrigger timeout counter 0.5secs


  return 0;
}


/* Cleanup - unregister the appropriate file from /proc */
void cleanup_module()
{
  // delete timeout counter
  del_timer_sync(&jiq_timer);

  kbdhook = NULL;
  printk("Cherry PS/2 kernel module is inactive now!\n");

  if(Device_Major == 0)
    printk("Error because Device_Major is 0\n");
}
