// ============================================================
//      
//      File : BitStream.cpp
//      
//      Copyright 1998 sepu gmbh
//                     Karlsruhe / Germany
//      
// ============================================================
//      




//------------------------------------------------------------
// Includes 
//------------------------------------------------------------
#include "bitstream.h"

#define NULL ((void * )0)

//#ifndef TEST

//#include <linux/config.h>
//#include <linux/kernel.h>
//#include <linux/sched.h>
//#include <linux/signal.h>
//#include <linux/errno.h>
//#include <linux/poll.h>
//#include <linux/init.h>
//#include <linux/slab.h>
//#include <linux/module.h>
//#include <linux/spinlock.h>
//#include <linux/list.h>
//#include <linux/smp_lock.h>
//#include <linux/usb.h>
//#include <linux/input.h>

//#endif

//#ifdef TEST

//#include <stdlib.h>
//#include <string.h>
//#include <fcntl.h>
//#include <unistd.h>
//#include <stdio.h>
//#include <sys/types.h>
//#include <sys/time.h>
//#include <sys/ioctl.h>
//#include <sys/signal.h>
//#include <asm/types.h>
//#include <errno.h>
//#include <syslog.h>

//#endif

extern BOOL CBitStream_LookAhead (BYTE * byNextBit);
extern USHORT CBitStream_m_uBitPos;
extern UINT CBitStream_m_uBytePos;
extern UINT CBitStream_m_uLen;
extern LPBYTE CBitStream_m_pBuffer;
extern BOOL CBitStream_m_fReference;	// no buffer copy

static struct
{
  BYTE *m_pBuffer;
  UINT m_uLen;
  UINT m_uBitPos;
  UINT m_uBytePos;
} CBitStreamData;


//------------------------------------------------------------
// Local Defines
//------------------------------------------------------------
const UINT BIT_MULTI = 8;
const UINT TOGGLE_OFFSET = 7;

//------------------------------------------------------------
// Static members
//------------------------------------------------------------

// ************************************************************
// * 
// * Construction/Destruction
// * 
// ************************************************************

//////////////////////////////////////////////////////////////
// Class    : CBitStream
// Function : Constructor
//      
//      Constructor of this class.
//      
//////////////////////////////////////////////////////////////
// 1. pBuffer    : pointer to byte buffer
// 2. uLen       : length of buffer
// 3. fReference : if true no copy is made 
//////////////////////////////////////////////////////////////

void
CBitStream_CBitStream (const BYTE * pBuffer, UINT uLen)
{
  CBitStreamData.m_pBuffer = NULL;
  CBitStreamData.m_uLen = uLen;
  CBitStreamData.m_uBitPos = 0;
  CBitStreamData.m_uBytePos = 0;
  CBitStreamData.m_pBuffer = (LPBYTE) pBuffer;
}

// ************************************************************
// * 
// * IMPLEMENTATION
// * 
// ************************************************************

//////////////////////////////////////////////////////////////
// Class    : CBitStream
// Function : GetNextBit
// Returns  : BYTE
//      
//      returns next bit of stream
//      
//////////////////////////////////////////////////////////////

BOOL
CBitStream_GetNextBit (BYTE * byNextBit)
{
  if (!CBitStreamData.m_pBuffer
      || CBitStreamData.m_uBytePos >= CBitStreamData.m_uLen)
    return FALSE;

  UINT nRealBitPos = 0;
  nRealBitPos = TOGGLE_OFFSET - CBitStreamData.m_uBitPos;
  *byNextBit =
    (CBitStreamData.
     m_pBuffer[CBitStreamData.m_uBytePos] & (1 << nRealBitPos)) ? 1 : 0;

  if (CBitStreamData.m_uBitPos == 7)
    {
      CBitStreamData.m_uBitPos = 0;
      CBitStreamData.m_uBytePos++;
    }
  else
    CBitStreamData.m_uBitPos++;

  return TRUE;
}

//////////////////////////////////////////////////////////////
// Class    : CBitStream
// Function : LookAhead
// Returns  : BYTE
//      
//      returns next bit of stream without removing it.
//      
//////////////////////////////////////////////////////////////

BOOL
CBitStream_LookAhead (BYTE * byNextBit)
{
  if (!CBitStreamData.m_pBuffer
      || CBitStreamData.m_uBytePos >= CBitStreamData.m_uLen)
    return FALSE;

  UINT nRealBitPos = TOGGLE_OFFSET - CBitStreamData.m_uBitPos;

  *byNextBit =
    (CBitStreamData.
     m_pBuffer[CBitStreamData.m_uBytePos] & (1 << nRealBitPos)) ? 1 : 0;
  return TRUE;
}

//////////////////////////////////////////////////////////////
// Class    : CBitStream
// Function : GetBitCount
// Returns  : UINT
//      
//      returns number of bits which are in the stream.
//      
//////////////////////////////////////////////////////////////

UINT
CBitStream_GetBitCount (void)
{
  if (CBitStreamData.m_uBytePos >= CBitStreamData.m_uLen)
    return 0;
  else
    return (CBitStreamData.m_uLen - CBitStreamData.m_uBytePos) * 8 + (8 -
								      CBitStreamData.
								      m_uBitPos);
}


// ************************************************************
// * 
// * Auxilary Functions
// * 
// ************************************************************

//////////////////////////////////////////////////////////////
// Class    : CBitStream
// Function : DeleteLeadingNulls
// Returns  : void
//      
//      delete the leading 0 from the bit stream.
//      
//////////////////////////////////////////////////////////////

void
CBitStream_DeleteLeadingNulls (void)
{
  BYTE byBit = 0;
  while (CBitStream_LookAhead (&byBit) && byBit == 0)
    CBitStream_GetNextBit (&byBit);
}
