// ============================================================
//      
//      File : Decode.cpp
//      
//      Copyright 2001 Cherry GmbH
//                     Auerbach / Germany
//      
// ============================================================
//      


//------------------------------------------------------------
// Local Defines
//------------------------------------------------------------
#define ISO_ALPHABET7	7
#define ISO_ALPHABET5	5



//------------------------------------------------------------
// Includes 
//------------------------------------------------------------



#include "bitstream.h"
#include "decode.h"

//#ifndef TEST

#include <linux/version.h>

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
#include <linux/config.h>
#else
#endif

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/smp_lock.h>
#include <linux/usb.h>
#include <linux/input.h>

//#endif

//#ifdef TEST

//#include <stdlib.h>
//#include <string.h>
//#include <fcntl.h>
//#include <unistd.h>
//#include <stdio.h>
//#include <sys/types.h>
//#include <sys/time.h>
//#include <sys/ioctl.h>
//#include <sys/signal.h>
//#include <asm/types.h>
//#include <errno.h>
//#include <syslog.h>

//#endif





//------------------------------------------------------------
// Static members
//------------------------------------------------------------
// translates

static OLECHAR ISO_Alphbet7[] = {
  ' ',
  '!',
  '"',
  '#',
  '$',
  '%',
  '&',
  0x27,				//'
  '(',
  ')',

  '*',
  '+',
  ',',
  '-',
  '.',
  '/',
  '0',
  '1',
  '2',
  '3',

  '4',
  '5',
  '6',
  '7',
  '8',
  '9',
  ':',
  ';',
  '<',
  '=',

  '>',
  '?',
  '@',
  'A',
  'B',
  'C',
  'D',
  'E',
  'F',
  'G',

  'H',
  'I',
  'J',
  'K',
  'L',
  'M',
  'N',
  'O',
  'P',
  'Q',

  'R',
  'S',
  'T',
  'U',
  'V',
  'W',
  'X',
  'Y',
  'Z',
  '[',

  '\\',
  ']',
  '^',
  '_'
};

static OLECHAR ISO_Alphbet5[] = {
  '0',
  '1',
  '2',
  '3',
  '4',
  '5',
  '6',
  '7',
  '8',
  '9',

  ':',
  ';',
  '<',
  '=',
  '>',
  '?'
};




typedef struct
{
  WORD m_wLength;
  BYTE m_bData[255];
} TrackData;




extern eCDIMSR_CARDERROR CDecode_DecodeISO (UINT auiNoBits,
					    const BYTE * apbTrack,
					    long alSize, BOOL abDecode);
extern eCDIMSR_CARDERROR CDecode_DecodeJIS2 (UINT auiTrack,
					     UINT auiNoBits,
					     const BYTE * apbTrack,
					     long alSize, BOOL abDecode);
extern void CDecode_ToAscii (OLECHAR * apTransTable, UINT alSize);
extern void CTrackData_Append (TrackData * CTrackData, BYTE bValue);
extern void CTrackData_Reset (TrackData * CTrackData);
extern WORD CTrackData_GetLength (TrackData * CTrackData);
extern void CTrackData_SetLength (TrackData * CTrackData, WORD wLength);
extern BOOL CTrackData_IsStopChar (TrackData * CTrackData, UINT nNoBits,
				   BYTE byRead);
extern BOOL CTrackData_IsStartChar (TrackData * CTrackData, UINT nNoBits,
				    BYTE byRead);
extern BOOL CTrackData_LRCErrorISO (TrackData * CTrackData, UINT nNoBits,
				    BYTE byLRC[7], BYTE cardLRC);
extern BOOL CTrackData_LRCErrorJIS2 (TrackData * CTrackData, UINT nNoBits,
				     BYTE byLRC[8], BYTE cardLRC);
extern void CTrackData_ToAscii (TrackData * CTrackData, OLECHAR * pTransTable,
				UINT alSize);




static struct
{
  UINT m_nStripe;
  UINT m_nStripeLength;
  BYTE *m_pInBuffer;
  BYTE *m_pbStartSentinel;
  BYTE *m_pbEndSentinel;
  BOOL m_abDecode;
  eCDIMSR_CARDTYPES *m_pCARD_TYPE;

  eCDIMSR_CARDTYPES m_eCardtype;
  BYTE m_bStartSentinel;
  BYTE m_bEndSentinel;
} CDecodeData;


TrackData m_Track;

// ************************************************************
// * 
// * CODE SECTION
// * 
// ************************************************************


void
CTrackData_Append (TrackData * CTrackData, BYTE bValue)
{
  if (CTrackData->m_wLength < 255)
    CTrackData->m_bData[CTrackData->m_wLength++] = bValue;
}

void
CTrackData_Reset (TrackData * CTrackData)
{
  CTrackData->m_wLength = 0;
}

WORD
CTrackData_GetLength (TrackData * CTrackData)
{
  return CTrackData->m_wLength;
}

void
CTrackData_SetLength (TrackData * CTrackData, WORD wLength)
{
  CTrackData->m_wLength = wLength;
}

// Decode.cpp: Implementierung der Klasse CDecode.
//
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////
// Class    : CDecode
// Function : Constructor
// Date     : 29.10.2001
// Author   : DreelH
// Descript : constructor
//      
//      Constructor of this class.
//////////////////////////////////////////////////////////////

void
CDecode_CDecode (void)
{
  CDecodeData.m_eCardtype = CARDTYPE_UNKNOWN;
  CDecodeData.m_bStartSentinel = 0;
  CDecodeData.m_bEndSentinel = 0;
  CTrackData_Reset (&m_Track);
}

//////////////////////////////////////////////////////////////
// Class    : CDecode
// Function : InterpreteSingleTrack
// Returns  : eCDIMSR_CARDERROR
// Date     : 29.10.2001
// Author   : DreelH
// Descript : interpret one track for ISO
//      
//////////////////////////////////////////////////////////////
// 1. nStripe         : 
// 2. nStripeLength   : 
// 3. pBuffer         : 
// 4. pbStartSentinel : 
// 5. pbEndSentinel   : 
//////////////////////////////////////////////////////////////

extern eCDIMSR_CARDERROR
CDecode_InterpreteSingleTrack (unsigned int nStripe,
			       unsigned int nStripeLength,
			       const char *pBuffer,
			       char *pbStartSentinel,
			       char *pbEndSentinel,
			       char abDecode, eCDIMSR_CARDTYPES * pCARD_TYPE)
//eCDIMSR_CARDERROR CDecode_InterpreteSingleTrack(UINT nStripe, 
//                                                                                                 UINT nStripeLength,
//                                                                                                       const BYTE * pBuffer,
//                                       BYTE* pbStartSentinel,
//                                       BYTE* pbEndSentinel,
//                                       BOOL abDecode,
//                                       eCDIMSR_CARDTYPES* pCARD_TYPE)
{
  eCDIMSR_CARDERROR eCardError;
  *pCARD_TYPE = CARDTYPE_UNKNOWN;
  UINT uiNoBits = nStripe == TRACK_1 ? 7 : 5;
  eCardError = CDecode_DecodeISO (uiNoBits, pBuffer, nStripeLength, abDecode);
  if (eCardError != MSR_S_OK)
    {
      if (nStripe == TRACK_2)
	{
	  // test if GEMINI card
	  uiNoBits = 7;
	  eCardError =
	    CDecode_DecodeISO (uiNoBits, pBuffer, nStripeLength, abDecode);
	}
      else if (nStripe == TRACK_3)
	{
	  // test if AAMVA card
	  uiNoBits = 7;
	  eCardError =
	    CDecode_DecodeISO (uiNoBits, pBuffer, nStripeLength, abDecode);
	}
    }
  if (eCardError != MSR_S_OK)
    {
      eCardError =
	CDecode_DecodeJIS2 (nStripe, uiNoBits, pBuffer, nStripeLength,
			    abDecode);
      uiNoBits = 8;
    }
  if (eCardError != MSR_S_OK)
    return eCardError;

  if (abDecode)
    {
      OLECHAR *pTransTable = NULL;
      UINT nSize = 0;
      if (uiNoBits && *pCARD_TYPE != CARDTYPE_JIS2)
	{
	  CDecodeData.m_eCardtype = CARDTYPE_ISO;
	  if (uiNoBits == 7)
	    {
	      pTransTable = ISO_Alphbet7;
	      nSize = sizeof (ISO_Alphbet7) / sizeof (OLECHAR);
	      if (nStripe == TRACK_2)
		CDecodeData.m_eCardtype = CARDTYPE_GEMINI;
	      else if (nStripe == TRACK_3)
		CDecodeData.m_eCardtype = CARDTYPE_AAMVA;
	    }
	  else
	    {
	      pTransTable = ISO_Alphbet5;
	      nSize = sizeof (ISO_Alphbet5) / sizeof (OLECHAR);
	    }
	}
      else
	CDecodeData.m_eCardtype = CARDTYPE_JIS2;

      CDecode_ToAscii (pTransTable, nSize);

      *pbStartSentinel = CDecodeData.m_bStartSentinel;
      *pbEndSentinel = CDecodeData.m_bEndSentinel;
    }
  *pCARD_TYPE = CDecodeData.m_eCardtype;
  return MSR_S_OK;
}

//////////////////////////////////////////////////////////////
// Class    : CDecode
// Function : IsStartChar
// Returns  : BOOL
// Date     : 29.10.2001
// Author   : DreelH
// Descript : test if byte is startsentinel
//      
//////////////////////////////////////////////////////////////
// 1. nNoBits : 
// 2. byRead  : 
//////////////////////////////////////////////////////////////

BOOL
CDecode_IsStartChar (UINT nNoBits, BYTE byRead)
{
  switch (nNoBits)
    {
    case 5:
      return byRead == 0xb;

    case 7:
      return byRead == 0x5;

    case 8:			// JIS 2
      return byRead == 0x7F;

    default:
      return FALSE;
    }
}

//////////////////////////////////////////////////////////////
// Class    : CDecode
// Function : IsStopChar
// Returns  : BOOL
// Date     : 29.10.2001
// Author   : DreelH
// Descript : test if byte is end sentinel
//      
//////////////////////////////////////////////////////////////
// 1. nNoBits : 
// 2. byRead  : 
//////////////////////////////////////////////////////////////

BOOL
CDecode_IsStopChar (UINT nNoBits, BYTE byRead)
{
  switch (nNoBits)
    {
    case 5:			// ISO
      return byRead == 0xf;

    case 7:			// ISO 
      return byRead == 0x1f;

    case 8:			// JIS2
      return byRead == 0x7f;

    default:
      return FALSE;
    }
}


//////////////////////////////////////////////////////////////
// Class    : CDecode
// Function : LRCError
// Returns  : BOOL
// Date     : 29.10.2001
// Author   : DreelH
// Descript : test LRC error
//      
//////////////////////////////////////////////////////////////
// 1. nNoBits  : 
// 2. byLRC[7] : 
// 3. cardLRC  : 
//////////////////////////////////////////////////////////////

BOOL
CDecode_LRCErrorISO (UINT nNoBits, BYTE byLRC[7], BYTE cardLRC)
{
  UINT nIndex;
  BYTE cardBit = 0;

  // while reading the stipe's content we added the bits
  // of the LRC to the byLRC array. This is the reason
  // why we have to substract the cardBit.
  for (nIndex = 0; nIndex < nNoBits; nIndex++)
    {
      cardBit = (cardLRC & (1 << nIndex)) ? 1 : 0;

      if ((byLRC[nIndex] % 2) != cardBit)
	return TRUE;
    }

  return FALSE;
}

BOOL
CDecode_LRCErrorJIS2 (UINT nNoBits, BYTE byLRC[8], BYTE cardLRC)
{
  UINT nIndex;
  BYTE cardBit = 0;

  // while reading the stipe's content we added the bits
  // of the LRC to the byLRC array. This is the reason
  // why we have to substract the cardBit.
  for (nIndex = 0; nIndex < nNoBits; nIndex++)
    {
      cardBit = (cardLRC & (1 << nIndex)) ? 1 : 0;

      if ((byLRC[nIndex] % 2) == cardBit)
	return TRUE;
    }

  return FALSE;
}

//////////////////////////////////////////////////////////////
// Class    : CDecode
// Function : GetTrackData
// Returns  : BOOL
// Date     : 29.10.2001
// Author   : DreelH
// Descript : read the interpreted track data
//      
//////////////////////////////////////////////////////////////
// 1. pbstrData : 
//////////////////////////////////////////////////////////////

int
CDecode_GetTrackData (BSTR * pbstrData)
{
  int i;
  int size;
  LPBYTE pv;
  size = CTrackData_GetLength (&m_Track);
  pv = (LPBYTE) pbstrData;
  for (i = 0; i < size; i++)
    pv[i] = m_Track.m_bData[i];
  return size;
}

void
CDecode_ToAscii (OLECHAR * apTransTable, UINT alSize)
{
  UINT nIndex;
  if (!CTrackData_GetLength (&m_Track))
    return;

  if (CDecodeData.m_eCardtype == CARDTYPE_JIS2)
    return;

  for (nIndex = 0; nIndex < (UINT) CTrackData_GetLength (&m_Track); nIndex++)
    {
      if (m_Track.m_bData[nIndex] >= alSize
	  || apTransTable[m_Track.m_bData[nIndex]] == 0)
	continue;		// not translateable

      BYTE b = apTransTable[m_Track.m_bData[nIndex]];
      m_Track.m_bData[nIndex] = b;

    }
  CDecodeData.m_bStartSentinel = apTransTable[CDecodeData.m_bStartSentinel];
}

//////////////////////////////////////////////////////////////
// Class    : CDecode
// Function : Decode
// Returns  : eCDIMSR_CARDERROR
// Date     : 21.10.2002
// Author   : dreelH
// Descript : decode if the data are ISO or JIS or ...
//      
//////////////////////////////////////////////////////////////
// 1. auiTrack  : 
// 2. auiNoBits : 
// 3. apbTrack  : 
// 4. alSize    : 
// 5. abDecode  : 
//////////////////////////////////////////////////////////////


eCDIMSR_CARDERROR
CDecode_DecodeJIS2 (UINT auiTrack,
		    UINT auiNoBits,
		    const BYTE * apbTrack, long alSize, BOOL abDecode)
{
  BYTE byLRC[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
  TrackData Track;
  CTrackData_Reset (&Track);

  UINT nNoBits = 8;
  BYTE byOneBit = 0, byOneByte = 0, byParity = 0, byParityMask = 0;
  BYTE byLRCByte = 0;
  UINT nBitPos = 0;
  BOOL fFoundStart = FALSE, fFoundEnd = FALSE, fFoundLRC = FALSE;
  CDecodeData.m_eCardtype = CARDTYPE_UNKNOWN;

  if (auiNoBits)
    nNoBits = auiNoBits;

  byParityMask = ~(1 << (nNoBits - 1));

  CBitStream_CBitStream (apbTrack, alSize);

  // delete all leading 0s
  CBitStream_DeleteLeadingNulls ();

  // is this an error condition?
  if (CBitStream_GetBitCount () == 0)
    return MSR_S_OK;

  long nBytes = 0;
  while (!fFoundLRC && CBitStream_GetNextBit (&byOneBit))
    {
      if (byOneBit == 1)
	{
	  if (!fFoundEnd)
	    {
	      // LRC 
	      // if fFoundEnd is true -> current byte is LRC byte ->
	      // do not add to the LRC array
	      byLRC[nBitPos]++;
	    }

	  // condition checks if the bit comes 
	  // fm the parity bit that is not included
	  // in the converted byte.
	  if (nBitPos + 1 != nNoBits)
	    {
	      byParity++;
	    }
	  // add parity bit to determinate 0 bytes
	  byOneByte += (1 << nBitPos);
	}
      nBitPos++;

      // one info. seq.
      if (nBitPos == nNoBits)
	{
	  nBytes++;
	  // ignore parity bit
	  if (!fFoundStart && nBytes == 1)
	    {
	      if (CDecode_IsStartChar (nNoBits, byOneByte & byParityMask))
		{
		  fFoundStart = TRUE;
		  nBitPos = 0;
		  byOneByte = 0;
		  byParity = 0;
		  continue;
		}
	    }

	  if (fFoundStart)
	    {
	      if (byOneByte == 0)
		{
		  // error - 0 byte can only occur at the end
		  //                 of stream
		  return MSR_E_FAILURE;
		}

	      // check parity (odd parity!!!!)
	      if ((byParity % 2) != byOneBit)
		{
		  // PARITY ERROR 
		  return MSR_E_PARITY;
		}

	      if (!fFoundEnd)
		{
		  // delete parity bit 
		  byOneByte &= byParityMask;
		  CTrackData_Append (&Track, byOneByte);

		  // parity bit already removed
		  if (CDecode_IsStopChar (nNoBits, byOneByte))
		    fFoundEnd = TRUE;
		}
	      else
		{
		  byLRCByte = byOneByte & byParityMask;
		  fFoundLRC = TRUE;
		}
	    }
	  nBitPos = 0;
	  byOneByte = 0;
	  byParity = 0;
	}
    }

  if (!fFoundStart)
    return MSR_E_START;
  if (!fFoundEnd)
    return MSR_E_END;

  if (!fFoundLRC)
    {
      // this case happens when the tail of the LRC
      // byte contains only nulls. Then it is possible
      // that the keyboard cuts these bits.
      // trailing zeros not sent
      byLRCByte = byOneByte;

      // paritiy is 0 -> number of bits must be even
      if ((byParity % 2) != 0)
	{
	  // PARITY ERROR 
	  return MSR_E_PARITY;
	}
    }

  // check track length
  if (CTrackData_GetLength (&Track) < 4)
    {
      // [start sentinal] [data - at least one byte] 
      // [end sentinal] [lrc]

      return MSR_E_LEN;
    }

  // LRC Check - last byte of the track data contains LRC
  // nNoBits - 1 because parity bit should not be tested
  if (CDecode_LRCErrorJIS2 (nNoBits - 1, byLRC, byLRCByte))
    return MSR_E_LRC;

  // COPY TRACK
  memcpy (m_Track.m_bData, Track.m_bData, CTrackData_GetLength (&Track) - 1);
  CTrackData_SetLength (&m_Track, CTrackData_GetLength (&Track) - 1);
  CDecodeData.m_bStartSentinel = 0x7F;
  CDecodeData.m_bEndSentinel = 0x7F;
  return MSR_S_OK;
}

eCDIMSR_CARDERROR
CDecode_DecodeISO (UINT auiNoBits,
		   const BYTE * apbTrack, long alSize, BOOL abDecode)
{
  BYTE byLRC[7] = { 0, 0, 0, 0, 0, 0, 0 };
  TrackData Track;
  CTrackData_Reset (&Track);

  UINT nNoBits = 0;
  BYTE byOneBit = 0, byOneByte = 0, byParity = 0, byParityMask = 0;
  BYTE byLRCByte = 0;
  UINT nBitPos = 0;
  BOOL fFoundStart = FALSE, fFoundEnd = FALSE, fFoundLRC = FALSE;
  CDecodeData.m_eCardtype = CARDTYPE_UNKNOWN;

  if (auiNoBits)
    nNoBits = auiNoBits;

  byParityMask = ~(1 << (nNoBits - 1));

  CBitStream_CBitStream (apbTrack, alSize);

  // delete all leading 0s
  CBitStream_DeleteLeadingNulls ();

  // is this an error condition?
  if (CBitStream_GetBitCount () == 0)
    {

      return MSR_S_OK;
    }

  OLECHAR *pTransTable = NULL;
  UINT nSize = 0;
  if (nNoBits == 7)
    {
      pTransTable = ISO_Alphbet7;
      nSize = sizeof (ISO_Alphbet7) / sizeof (OLECHAR);
    }
  else
    {
      pTransTable = ISO_Alphbet5;
      nSize = sizeof (ISO_Alphbet5) / sizeof (OLECHAR);
    }

  while (!fFoundLRC && CBitStream_GetNextBit (&byOneBit))
    {
      if (byOneBit == 1)
	{
	  if (!fFoundEnd)
	    {
	      // LRC 
	      // if fFoundEnd is true -> current byte is LRC byte ->
	      // do not add to the LRC array
	      byLRC[nBitPos]++;
	    }

	  // condition checks if the bit comes 
	  // fm the parity bit that is not included
	  // in the converted byte.
	  if (nBitPos + 1 != nNoBits)
	    {
	      byParity++;
	    }
	  // add parity bit to determinate 0 bytes
	  byOneByte += (1 << nBitPos);
	}
      nBitPos++;

      // one info. seq.
      if (nBitPos == nNoBits)
	{
	  // ignore parity bit
	  if (!fFoundStart)
	    fFoundStart =
	      CDecode_IsStartChar (nNoBits, byOneByte & byParityMask);

	  if (fFoundStart)
	    {
	      if (byOneByte == 0)
		{
		  // error - 0 byte can only occur at the end
		  //                 of stream
		  return MSR_E_FAILURE;
		}

	      // check parity (odd parity!!!!)
	      if ((byParity % 2) == byOneBit)
		{
		  // PARITY ERROR 
		  return MSR_E_PARITY;
		}

	      if (!fFoundEnd)
		{
		  // delete parity bit 
		  byOneByte &= byParityMask;
		  CTrackData_Append (&Track, byOneByte);

		  // parity bit already removed
		  if (CDecode_IsStopChar (nNoBits, byOneByte))
		    {
		      fFoundEnd = TRUE;
		      if (abDecode && byOneByte < nSize)
			CDecodeData.m_bEndSentinel = pTransTable[byOneByte];
		      else
			CDecodeData.m_bEndSentinel = byOneByte;
		    }
		}
	      else
		{
		  byLRCByte = byOneByte & byParityMask;
		  fFoundLRC = TRUE;
		}
	    }
	  nBitPos = 0;
	  byOneByte = 0;
	  byParity = 0;
	}
    }

  if (!fFoundStart)
    return MSR_E_START;
  if (!fFoundEnd)
    return MSR_E_END;

  if (!fFoundLRC)
    {
      // this case happens when the tail of the LRC
      // byte contains only nulls. Then it is possible
      // that the keyboard cuts these bits.
      // trailing zeros not sent
      byLRCByte = byOneByte;

      // paritiy is 0 -> number of bits must be odd
      if ((byParity % 2) == 0)
	{
	  // PARITY ERROR 
	  return MSR_E_PARITY;
	}
    }

  // check track length
  if (CTrackData_GetLength (&Track) < 4)
    {
      // [start sentinal] [data - at least one byte] 
      // [end sentinal] [lrc]

      return MSR_E_LEN;
    }

  // LRC Check - last byte of the track data contains LRC
  // nNoBits - 1 because parity bit should not be tested
  if (CDecode_LRCErrorISO (nNoBits - 1, byLRC, byLRCByte))
    return MSR_E_LRC;

  // COPY TRACK
  memcpy (m_Track.m_bData, Track.m_bData + 1,	// start sentinal 
	  CTrackData_GetLength (&Track) - 2);	// start sentinal + end sentinal
  CTrackData_SetLength (&m_Track, CTrackData_GetLength (&Track) - 2);
  CDecodeData.m_bStartSentinel = Track.m_bData[0];
  return MSR_S_OK;
}
