
/*
 * ckbd_drv.c Interface to Cherry Keyboards 
 */

/*
 * The necessary header files 
 */

/*
 * Standard in kernel modules 
 */
#include <linux/kernel.h>	/* We're doing kernel work */
#include <linux/module.h>	/* Specifically, a module */

#include <linux/slab.h>
#include <linux/input.h>	// um Keys weiterzuleiten

#include <asm/uaccess.h>	/* for get_user and put_user */

#include <linux/sched.h>
/*
 * The character device definitions are here 
 */
#include <linux/fs.h>

/*
 * Our own definitions 
 */
#include "msr_devs.h"
#include "decode.h"
#include "raw2kbd.h"
#include "scancodes.h"

/*
 * to output some debug messages 
 */

#define MY_DEBUG

/*
 * Defines 
 */

#define STAT_LEN 8

#define TR1START 3
#define TR2START 90
#define TR3START 139

/*
 * Device Declarations ******************************** 
 */

/*
 * for timeout timer 
 */
struct timer_list jpos_timer;

/*
 * Device Declarations ******************************** 
 */

/*
 * The name for our device, as it will appear in /proc/devices 
 */
#define DEVICE_NAME "chyposusbmsr"

/*
 * The maximum length of the message for the device 
 */
#define BUF_LEN 256

/*
 * Is the device open right now? Used to prevent concurent access into the 
 * same device 
 */
static int Device_Open = 0;
static int Device_Minor = 0;
static int Device_Major = 0;

static spinlock_t ckbd_u_lock;
static int ckbd_u_count = 0;
static unsigned int ckbd_u_owner = 0;

static int ckbd_u_stat = 0;
static int ckbd_u_msr1 = 0;
static int ckbd_u_msr2 = 0;
static int ckbd_u_msr3 = 0;
static int ckbd_u_keylock = 0;

/*
 * The message the device will give when asked 
 */
static char Message[BUF_LEN];

/*
 * How far did the process reading the message get? Useful if the message
 * is larger than the size of the buffer we get to fill in device_read. 
 */
static char *Message_Ptr;
static size_t Message_Len = 0;

static unsigned char msrold[9];

static unsigned char usb_msr_keycode[256] = {
  0, 0, 0, 0, 30, 48, 46, 32, 18, 33, 34, 35, 23, 36, 37, 38,
  50, 49, 24, 25, 16, 19, 31, 20, 22, 47, 17, 45, 21, 44, 2, 3,
  4, 5, 6, 7, 8, 9, 10, 11, 28, 1, 14, 15, 57, 12, 13, 26,
  27, 43, 43, 39, 40, 41, 51, 52, 53, 58, 59, 60, 61, 62, 63, 64,
  65, 66, 67, 68, 87, 88, 99, 70, 119, 110, 102, 104, 111, 107, 109, 106,
  105, 108, 103, 69, 98, 55, 74, 78, 96, 79, 80, 81, 75, 76, 77, 71,
  72, 73, 82, 83, 86, 127, 116, 117, 183, 184, 185, 186, 187, 188, 189, 190,
  191, 192, 193, 194, 134, 138, 130, 132, 128, 129, 131, 137, 133, 135, 136,
  113,
  115, 114, 0, 0, 0, 121, 0, 89, 93, 124, 92, 94, 95, 0, 0, 0,
  122, 123, 90, 91, 85, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  29, 42, 56, 125, 97, 54, 100, 126, 164, 166, 165, 163, 161, 115, 114, 113,
  150, 158, 159, 128, 136, 177, 178, 176, 142, 152, 173, 140
};



/*
 * Stat data area 
 */
static struct
{
  short msr1status;
  short msr2status;
  short msr3status;
  unsigned char keylockposition;
} StatData;

static int STAT_len = 0;
static unsigned char *pStatData;

#define MSR1_STATUS  StatData.msr1status
#define MSR2_STATUS  StatData.msr2status
#define MSR3_STATUS  StatData.msr3status
#define KLPOS  StatData.keylockposition

/*
 * MSR data area 
 */
static unsigned char MSR1Data[BUF_LEN];
static unsigned char *pMSR1Data;
static unsigned char MSR2Data[BUF_LEN];
static unsigned char *pMSR2Data;
static unsigned char MSR3Data[BUF_LEN];
static unsigned char *pMSR3Data;

static unsigned char MSRData[512];
static unsigned char MODData[512];
static unsigned int MSRLen;
static unsigned char MSRState;
static unsigned int MSR1Len;
static unsigned int MSR2Len;
static unsigned int MSR3Len;
static unsigned char msrerror;
static unsigned char msrerrorcnt;

/*
 * Header und Terminatorkonfiguration fuer MSR 
 */

static unsigned char termconfig[17];

#define BuzzLen termconfig[0]
#define GHmod   termconfig[1]
#define GHcode  termconfig[2]
#define GTmod   termconfig[3]
#define GTcode  termconfig[4]

#define H1mod   termconfig[5]
#define H1code  termconfig[6]
#define T1mod   termconfig[7]
#define T1code  termconfig[8]

#define H2mod   termconfig[9]
#define H2code  termconfig[10]
#define T2mod   termconfig[11]
#define T2code  termconfig[12]

#define H3mod   termconfig[13]
#define H3code  termconfig[14]
#define T3mod   termconfig[15]
#define T3code  termconfig[16]

static unsigned char *pKeylockData;

/*
 * This function is called whenever a process attempts to open the device
 * file 
 */
static int
device_open (struct inode *inode, struct file *file)
{
#ifdef MY_DEBUG
  printk ("device_open(%p)\n", file);
#endif

  spin_lock (&ckbd_u_lock);
  if (ckbd_u_count && (ckbd_u_owner != current->uid) &&	/* allow user */
      (ckbd_u_owner != current->euid) &&	/* allow whoever did su */
      !capable (CAP_DAC_OVERRIDE))
    {				/* still allow root */
      spin_unlock (&ckbd_u_lock);
      return -EBUSY;
    }

  if (ckbd_u_count == 0)
    {
      ckbd_u_owner = current->uid;	/* grab it */
    }

  ckbd_u_count++;
  spin_unlock (&ckbd_u_lock);

  /*
   * If this was a process, we would have had to be more careful here,
   * because one process might have checked Device_Open right before the 
   * other one tried to increment it. However, we're in the kernel, so
   * we're protected against context switches. This is NOT the right
   * attitude to take, because we might be running on an SMP box, but
   * we'll deal with SMP in a later chapter. 
   */

  Device_Open++;
  Device_Minor = MINOR (inode->i_rdev);
#ifdef MY_DEBUG
  printk ("device_open minor = %d\n", Device_Minor);
#endif
  switch (Device_Minor)
    {
    case STAT_MINOR_NUM:
      ckbd_u_stat++;
      break;
    case MSR1_MINOR_NUM:
      ckbd_u_msr1++;
      pMSR1Data = &MSR1Data[0];
      MSR1_STATUS = 0;
      break;
    case MSR2_MINOR_NUM:
      ckbd_u_msr2++;
      pMSR2Data = &MSR2Data[0];
      MSR2_STATUS = 0;
      break;
    case MSR3_MINOR_NUM:
      ckbd_u_msr3++;
      pMSR3Data = &MSR3Data[0];
      MSR3_STATUS = 0;
      break;
    case KEYLOCK_MINOR_NUM:
      ckbd_u_keylock++;
      break;
    }
  /*
   * Initialize the message 
   */
  Message_Ptr = Message;
  Message_Len = 0;

  return 0;
}

/*
 * This function is called when a process closes the device file. It
 * doesn't have a return value because it cannot fail. Regardless of what
 * else happens, you should always be able to close a device (in 2.0, a
 * 2.2 device file could be impossible to close). 
 */
static int
device_release (struct inode *inode, struct file *file)
{
#ifdef MY_DEBUG
  printk ("device_release(%p,%p)\n", inode, file);
#endif
  Device_Minor = MINOR (inode->i_rdev);
#ifdef MY_DEBUG
  printk ("device_release minor = %d\n", Device_Minor);
#endif
  switch (Device_Minor)
    {
    case STAT_MINOR_NUM:
      ckbd_u_stat--;
      break;
    case MSR1_MINOR_NUM:
      ckbd_u_msr1--;
      break;
    case MSR2_MINOR_NUM:
      ckbd_u_msr2--;
      break;
    case MSR3_MINOR_NUM:
      ckbd_u_msr3--;
      break;
    case KEYLOCK_MINOR_NUM:
      ckbd_u_keylock--;
      break;
    }

  /*
   * We're now ready for our next caller 
   */
  Device_Open--;
  ckbd_u_count--;

  return 0;
}

/*
 * This function is called whenever a process which has already opened the 
 * device file attempts to read from it. 
 */

static ssize_t
device_read (struct file *file, char *buffer,	/* The buffer to fill with 
						 * the data */
	     size_t length,	/* The length of the buffer */
	     loff_t * offset)
{				/* offset to the file */
  int bytes_read = 0;

  Device_Minor = MINOR (file->f_dentry->d_inode->i_rdev);

#ifdef MY_DEBUG
  printk ("device_read(%d,%p,%p,%d)\n", Device_Minor, file, buffer, length);
#endif

  /*
   * Number of bytes actually written to the buffer 
   */
  if (length <= 0)
    return -EINVAL;

  if (Device_Minor == STAT_MINOR_NUM)
    {				// STAT
#ifdef MY_DEBUG
      printk ("Read STAT data %d\n", STAT_len);
#endif
      if (length > BUF_LEN)
	length = BUF_LEN;
      pStatData = (char *) &StatData;
      Message_Ptr = pStatData;	// StatData
      STAT_len = STAT_LEN;
    }
  else if (Device_Minor == MSR1_MINOR_NUM)
    {				// KEYBOARD
#ifdef MY_DEBUG
      printk ("Read MSR1 data %d\n", length);
#endif
      if (length > BUF_LEN)
	length = BUF_LEN;
      pMSR1Data = &MSR1Data[0];
      Message_Ptr = pMSR1Data;
    }
  else if (Device_Minor == MSR2_MINOR_NUM)
    {				// KEYBOARD
#ifdef MY_DEBUG
      printk ("Read MSR2 data %d\n", length);
#endif
      if (length > BUF_LEN)
	length = BUF_LEN;
      pMSR2Data = &MSR2Data[0];
      Message_Ptr = pMSR2Data;
    }
  else if (Device_Minor == MSR3_MINOR_NUM)
    {				// KEYBOARD
#ifdef MY_DEBUG
      printk ("Read MSR3 data %d\n", length);
#endif
      if (length > BUF_LEN)
	length = BUF_LEN;
      pMSR3Data = &MSR3Data[0];
      Message_Ptr = pMSR3Data;
    }
  if (Device_Minor == KEYLOCK_MINOR_NUM)
    {				// KEYLOCK
#ifdef MY_DEBUG
      printk ("Read Keylock data \n");
#endif
      if (length > BUF_LEN)
	length = BUF_LEN;
      pKeylockData = (char *) &KLPOS;
      Message_Ptr = pKeylockData;	// StatData
    }

  /*
   * If we're at the end of the message, return 0 (which signifies end
   * of file) 
   */
  if (Message_Ptr == 0)
    return 0;

  /*
   * Actually put the data into the buffer 
   */
#ifdef MY_DEBUG
  printk ("Read data: ");
#endif
  while (length)
    {

      /*
       * Because the buffer is in the user data segment, not the kernel
       * data segment, assignment wouldn't work. Instead, we have to use 
       * put_user which copies data from the kernel data segment to the
       * user data segment. 
       */
#ifdef MY_DEBUG
      printk ("%x ", *Message_Ptr);
#endif
      put_user (*(Message_Ptr++), buffer++);
      length--;
      bytes_read++;
    }
#ifdef MY_DEBUG
  printk ("\nRead %d bytes, %d left\n", bytes_read, length);
#endif

  /*
   * ReInitialize the message 
   */

  if (Device_Minor == MSR1_MINOR_NUM)
    {
      pMSR1Data = &MSR1Data[0];
      MSR1_STATUS = 0;
    }
  if (Device_Minor == MSR2_MINOR_NUM)
    {
      pMSR2Data = &MSR2Data[0];
      MSR2_STATUS = 0;
    }
  if (Device_Minor == MSR3_MINOR_NUM)
    {
      pMSR3Data = &MSR3Data[0];
      MSR3_STATUS = 0;
    }

  /*
   * Read functions are supposed to return the number of bytes actually
   * inserted into the buffer 
   */
  return bytes_read;
}

/*
 * This function is called when somebody tries to write into our device
 * file. 
 */
static ssize_t
device_write (struct file *file,
	      const char *buffer, size_t length, loff_t * offset)
{
  unsigned int i = 0;

  Device_Minor = MINOR (file->f_dentry->d_inode->i_rdev);

  /** Write to STAT
  *	send buffer to keyboard firmware
  */
  if (Device_Minor == STAT_MINOR_NUM)
    {				// STAT
      for (i = 0; (i < length) && (i < sizeof (termconfig)); i++)
	{
	  get_user (termconfig[i], buffer + i);
#ifdef MY_DEBUG
	  printk ("Terminatorconfig[%d] = %x\n", i, termconfig[i]);
#endif
	}
    }

  /** Write to other devices
  *
  */
  else
    {
      printk ("device_write not available for this device %d\n",
	      Device_Minor);
      i = 0;
    }

  Message_Ptr = Message;

  /*
   * Again, return the number of input characters used 
   */
  return i;
}

void reinit_jpos (unsigned long ptr)
{
  MSRLen = 0;
  MSRState = 0;
  MSR1Len = 0;
  MSR2Len = 0;
  MSR3Len = 0;
}

void
msrout ()
{
#ifdef MY_DEBUG
  int i;

  printk ("MSR1: ");
  for (i = 0; i < MSR1Len; i++)
    {
      printk ("%c", MSR1Data[i]);
    }
  printk ("\n");
  printk ("MSR2: ");
  for (i = 0; i < MSR2Len; i++)
    {
      printk ("%c", MSR2Data[i]);
    }
  printk ("\n");
  printk ("MSR3: ");
  for (i = 0; i < MSR3Len; i++)
    {
      printk ("%c", MSR3Data[i]);
    }
  printk ("\n");
  printk ("Error = %x\n", msrerror);
#endif
  if (msrerror == 0)
    {
      MSR1Data[0] = MSR_E_GENERAL;
      MSR1Len = 1;
      MSR2Data[0] = MSR_E_GENERAL;
      MSR2Len = 1;
      MSR3Data[0] = MSR_E_GENERAL;
      MSR3Len = 1;
    }
  if (msrerror & 1)
    {				/* track error */
      MSR1Data[0] = MSR_E_FAILURE;
      MSR1Len = 1;
    }
  if (msrerror & 2)
    {				/* track error */
      MSR2Data[0] = MSR_E_FAILURE;
      MSR2Len = 1;
    }
  if (msrerror & 4)
    {				/* track error */
      MSR3Data[0] = MSR_E_FAILURE;
      MSR3Len = 1;
    }

  MSR1_STATUS = MSR1Len;	/* signal no of bytes to read */
  MSR2_STATUS = MSR2Len;	/* signal no of bytes to read */
  MSR3_STATUS = MSR3Len;	/* signal no of bytes to read */

}

char
hid2bin (char hidwert)
{
  switch (hidwert)
    {
    case 0x27:			/* "0" */
      return 0;
      break;
    case 0x1e:			/* "1" */
      return 1;
      break;
    case 0x1f:			/* "2" */
      return 2;
      break;
    case 0x20:			/* "3" */
      return 3;
      break;
    case 0x21:			/* "4" */
      return 4;
      break;
    case 0x22:			/* "5" */
      return 5;
      break;
    case 0x23:			/* "6" */
      return 6;
      break;
    case 0x24:			/* "7" */
      return 7;
      break;
    default:
      return 0x0f;
    }
}

/*
 * Interpret Magnetic Stripes 
 */
unsigned char
msrcheck (struct input_dev *idev, unsigned char *msr_data,
	  unsigned char apusb)
{

  /*
   * Aufbau msr Puffer [256] buffer[0] = length track 1 buffer[1] =
   * length track 2 buffer[2] = length track 3 buffer[3] = start track 1
   * buffer[90] = start track 2 buffer[139] = start track 3 
   */
  unsigned char reversetrack[BUF_LEN];
  int result;
  char pCARD_TYPE;
  char pbStartSentinel;
  char pbEndSentinel;
  int retval;

  
  retval = BuzzLen;

#ifdef MY_DEBUG
  printk ("msrcheck\n");
#endif

  if (apusb)
    {

      if (ckbd_u_stat)
	{			/* msr status open open dont send input to 
				 * keyboard */

	  CDecode_CDecode ();

	  if (msr_data[0])
	    {			/* track 1 has data */

#ifdef MY_DEBUG
	      printk ("msr1 laenge = %x\n", msr_data[0]);
#endif

	      inverttrack (&msr_data[TR1START], &reversetrack[0],
			   msr_data[0]);
	      result =
		CDecode_InterpreteSingleTrack (1, msr_data[0],
					       &reversetrack[0],
					       &pbStartSentinel,
					       &pbEndSentinel, 1,
					       &pCARD_TYPE);
	      if (result != MSR_S_OK)
		result =
		  CDecode_InterpreteSingleTrack (1, msr_data[0],
						 &msr_data[TR1START],
						 &pbStartSentinel,
						 &pbEndSentinel, 1,
						 &pCARD_TYPE);
	      if (result == MSR_S_OK)
		{
		  result = CDecode_GetTrackData (&MSR1Data[2]);
		  if (result)
		    {		/* prepare the result */
		      MSR1Data[0] = MSR_S_OK;	/* status decoding */
		      MSR1Data[1] = pbStartSentinel;
		      MSR1Data[result + 2] = pbEndSentinel;
		      MSR1_STATUS = result + 3;	/* signal no of
						 * bytes to read */
		    }
		  else
		    {		/* Cecoding failed, signal error */
		      MSR1Data[0] = 0xff;	/* error */
		      MSR1_STATUS = 1;	/* one byte to read */
		      retval = 0;
		    }
		}
	      else
		{
		  MSR1Data[0] = result;	/* errorcode */
		  MSR1_STATUS = 1;	/* one byte to read */
		  retval = 0;
		}
	    }

	  if (msr_data[1])
	    {

#ifdef MY_DEBUG
	      printk ("msr2 laenge = %x\n", msr_data[1]);
#endif
	      inverttrack (&msr_data[TR2START], &reversetrack[0],
			   msr_data[1]);
	      result =
		CDecode_InterpreteSingleTrack (2, msr_data[1],
					       &reversetrack[0],
					       &pbStartSentinel,
					       &pbEndSentinel, 1,
					       &pCARD_TYPE);
	      if (result != MSR_S_OK)
		result =
		  CDecode_InterpreteSingleTrack (2, msr_data[1],
						 &msr_data[TR2START],
						 &pbStartSentinel,
						 &pbEndSentinel, 1,
						 &pCARD_TYPE);
	      if (result == MSR_S_OK)
		{
		  result = CDecode_GetTrackData (&MSR2Data[2]);
		  if (result)
		    {
		      MSR2Data[0] = MSR_S_OK;
		      MSR2Data[1] = pbStartSentinel;
		      MSR2Data[result + 2] = pbEndSentinel;
		      MSR2_STATUS = result + 3;
		    }
		  else
		    {
		      MSR2Data[0] = 0xff;
		      MSR2_STATUS = 1;
		      retval = 0;
		    }
		}
	      else
		{
		  MSR2Data[0] = result;
		  MSR2_STATUS = 1;
		  retval = 0;
		}
	    }

	  if (msr_data[2])
	    {

#ifdef MY_DEBUG
	      printk ("msr3 laenge = %x\n", msr_data[2]);
#endif
	      inverttrack (&msr_data[TR3START], &reversetrack[0],
			   msr_data[2]);
	      result =
		CDecode_InterpreteSingleTrack (3, msr_data[2],
					       &reversetrack[0],
					       &pbStartSentinel,
					       &pbEndSentinel, 1,
					       &pCARD_TYPE);
	      if (result != MSR_S_OK)
		result =
		  CDecode_InterpreteSingleTrack (3, msr_data[2],
						 &msr_data[TR3START],
						 &pbStartSentinel,
						 &pbEndSentinel, 1,
						 &pCARD_TYPE);
	      if (result == MSR_S_OK)
		{
		  result = CDecode_GetTrackData (&MSR3Data[2]);
		  if (result)
		    {
		      MSR3Data[0] = MSR_S_OK;
		      MSR3Data[1] = pbStartSentinel;
		      MSR3Data[result + 2] = pbEndSentinel;
		      MSR3_STATUS = result + 3;
		    }
		  else
		    {
		      MSR3Data[0] = 0xff;
		      MSR3_STATUS = 1;
		      retval = 0;
		    }
		}
	      else
		{
		  MSR3Data[0] = result;
		  MSR3_STATUS = 1;
		  retval = 0;
		}
	    }
	}
      else
	{
	  raw2kbd (idev, msr_data);
	}

    }
  else
    {				/* smallpos */

      unsigned char mod;
      unsigned char code;
      int i;

#define WAITFORGLOBALHEADER 0
#define WAITFORMSRHEADER 1
#define RECMSR1	2
#define RECMSR2	3
#define RECMSR3	4


#ifdef MY_DEBUG
      printk ("smallpos keyboard\n");
#endif


      if (ckbd_u_stat)
	{			/* msr status open open dont send input to keyboard */

          // retrigger timeout counter
          mod_timer (&jpos_timer, jiffies + HZ/2);	// retrigger timeout counter  0.5secs
	  for (i = 3; i < 9; i++)
	    {

	      if (msr_data[i] > 3
		  && memscan (&msrold[0] + 3, msr_data[i],
			      6) == &msrold[0] + 9)
		{

		  mod = msr_data[1];
		  code = msr_data[i];

		  // mod = msr_data[0];
		  // code = msr_data[1];

		  MSRData[MSRLen] = code;
		  MODData[MSRLen] = mod;

		  if (MSRLen < 511)
		    MSRLen++;

		  switch (MSRState)
		    {
		    case WAITFORMSRHEADER:	/* Wait for MSR 1 header */
#ifdef MY_DEBUG
		      printk ("Wait for MSR header\n");
#endif
		      if ((mod == H1mod) && (code == H1code))
			{
			  MSRState = RECMSR1;
			  MSR1Len++;
			}
		      if ((mod == H2mod) && (code == H2code))
			{
			  MSRState = RECMSR2;
			  MSR2Len++;
			}
		      if ((mod == H3mod) && (code == H3code))
			{
			  MSRState = RECMSR3;
			  MSR3Len++;
			}
		      if ((mod == GTmod) && (code == GTcode))
			{
			  msrout ();
			  MSRState = WAITFORGLOBALHEADER;
			}
		      if (MSRState == WAITFORMSRHEADER)
			{
			  if (msrerrorcnt)
			    {
			      msrerror |= hid2bin (code);
			    }
			  else
			    {
			      msrerror = (hid2bin (code) << 4);
			      msrerrorcnt = 1;
			    }
			}
		      break;

		    case RECMSR1:	/* Receive MSR 1 data */
#ifdef MY_DEBUG
		      printk ("Receive MSR 1 data, mod=0x%08X, code=0x%08X\n",mod,code);
#endif
		      if ((mod == T1mod) && (code == T1code))
			{	/* check for msr1 terminator */
			  MSRState = WAITFORMSRHEADER;
			}
		      else
			{	/* write msr1 data */
			  if ((mod & 0x22) != 0)
			    {
			      MSR1Data[MSR1Len] = usb2ascsft[code];
			    }
			  else if ((mod & 0x40) != 0)
			    {
			      MSR1Data[MSR1Len] = usb2ascAlt[code];
			    }
			  else
			    {
			      MSR1Data[MSR1Len] = usb2ascstd[code];
			    }
			  if (MSR1Len < BUF_LEN - 1)
			    MSR1Len++;
			}
		      break;

		    case RECMSR2:	/* Receive MSR 2 data */
#ifdef MY_DEBUG
		      printk ("Receive MSR 2 data\n");
#endif
		      if ((mod == T2mod) && (code == T2code))
			{	/* check for msr2 terminator */
			  MSRState = WAITFORMSRHEADER;
			}
		      else
			{	/* write msr2 data */
			  if ((mod & 0x22) != 0)
			    {
			      MSR2Data[MSR2Len] = usb2ascsft[code];
			    }
			  else if ((mod & 0x40) != 0)
			    {
			      MSR2Data[MSR2Len] = usb2ascAlt[code];
			    }
			  else 

			    {
			      MSR2Data[MSR2Len] = usb2ascstd[code];
			    }
			  if (MSR2Len < BUF_LEN - 1)
			    MSR2Len++;
			}
		      break;

		    case RECMSR3:	/* Receive MSR 3 data */
#ifdef MY_DEBUG
		      printk ("Receive MSR 3 data\n");
#endif
		      if ((mod == T3mod) && (code == T3code))
			{	/* check for msr3 terminator */
			  MSRState = WAITFORMSRHEADER;
			}
		      else
			{	/* write msr3 data */
			  if ((mod & 0x22) != 0)
			    {
			      MSR3Data[MSR3Len] = usb2ascsft[code];
			    }
			  else if ((mod & 0x40) != 0)
			    {
			      MSR3Data[MSR3Len] = usb2ascAlt[code];
			    }
			  else
			    {
			      MSR3Data[MSR3Len] = usb2ascstd[code];
			    }
			  if (MSR3Len < BUF_LEN - 1)
			    MSR3Len++;
			}
		      break;

		    default:	/* 0 or error, wait for global header */
#ifdef MY_DEBUG
		      printk ("Wait for Global Header\n");
#endif
		      MSRState = WAITFORGLOBALHEADER;
		      MSR1Len = 0;
		      MSR2Len = 0;
		      MSR3Len = 0;
		      MSR1Data[0] = MSR_S_OK;
		      MSR2Data[0] = MSR_S_OK;
		      MSR3Data[0] = MSR_S_OK;
		      msrerrorcnt = 0;
		      if ((mod == GHmod) && (code == GHcode))
			{
			  MSRState = WAITFORMSRHEADER;
			}
		    }

		}
	    }


	}
      else
	{			/* output delayed on console */

#ifdef MY_DEBUG
	  printk ("buffered output\n");
	  printk ("daten :");
	  for (i = 0; i < 9; i++)
	    {
	      printk (" %x", msr_data[i]);
	    }
	  printk ("\n");
	  printk ("old data :");
	  for (i = 0; i < 9; i++)
	    {
	      printk (" %x", msrold[i]);
	    }
	  printk ("\n");
#endif


/*
the original linux keyboard code
*	  for (i = 0; i < 8; i++)
*	    bufferkbd (idev, usb_msr_keycode[i + 224],
*		       (msr_data[1] >> i) & 1);
would lead to at least 8 events per keypress
(8 modifier events are always generated as up events if no modifier key is pressed.
*/

/* To minimize overhead for delayed output the following code is used.*/


	  for (i = 0; i < 8; i++)
	    {
	      if (((msr_data[1] >> i) & 1) != ((msrold[1] >> i) & 1))
		{
		  intobuffer (idev, (int) (usb_msr_keycode[i + 224]) + ((int) ((msr_data[1] >> i) & 1) << 8));
		}
	    }

/* the rest is the original linux code again */

	  for (i = 3; i < 9; i++)
	    {

	      if (msrold[i] > 3
		  && memscan (&msr_data[0] + 3, msrold[i],
			      6) == &msr_data[0] + 9)
		{
		  if (usb_msr_keycode[msrold[i]])
		    {
		      intobuffer (idev, (int) usb_msr_keycode[msrold[i]] + UP);
		    }
		}

	      if (msr_data[i] > 3
		  && memscan (&msrold[0] + 3, msr_data[i],
			      6) == &msrold[0] + 9)
		{
		  if (usb_msr_keycode[msr_data[i]])
		    {
		      intobuffer (idev, usb_msr_keycode[msr_data[i]] + DOWN);
		    }
		}
	    }

	}
      memcpy (&msrold[0], &msr_data[0], 9);
    }
  return retval;
}

void
keylock (struct input_dev *idev, unsigned char position)
{

  if (position & 0x01)
    {				/* keylock position 1 */
      KLPOS = 1;
    }
  else if (position & 0x02)
    {				/* keylock position 2 */
      KLPOS = 2;
    }
  else if (position & 0x04)
    {				/* keylock position 3 */
      KLPOS = 3;
    }
  else if (position & 0x08)
    {				/* keylock position 4 */
      KLPOS = 4;
    }
  else if (position & 0x10)
    {				/* keylock position 5 */
      KLPOS = 5;
    }
  else
    {				/* fehler */
      KLPOS = 0;
    }

#ifdef MY_DEBUG
  printk ("Keylock position %x received\n", KLPOS);
#endif
}

/*
 * This structure will hold the functions to be called when a process does 
 * something to the device we created. Since a pointer to this structure
 * is kept in the devices table, it can't be local to init_module. NULL is 
 * for unimplemented functions. 
 */
struct file_operations Fops = {
owner:THIS_MODULE,
read:device_read,
write:device_write,
open:device_open,
release:device_release
};

/*
 * Initialize and Register the character device 
 */
int
init_msr ()
{
  int ret_val;

  /*
   * Register the character device (at least try) 
   */
  ret_val = register_chrdev (CKBD_MAJOR_NUM, DEVICE_NAME, &Fops);

  /*
   * Negative values signify an error 
   */
  if (ret_val < 0)
    {
      printk ("Registering the character device failed with %d\n", ret_val);
      return ret_val;
    }
  Device_Major = ret_val;

  printk ("chyposusbmsr registered with major device number %d.\n",
	  Device_Major);

  spin_lock_init (&ckbd_u_lock);	// init spin_lock

  /*
   * init STAT data 
   */
  StatData.msr1status = 0;
  StatData.msr2status = 0;
  StatData.msr3status = 0;
  STAT_len = 0;
  pStatData = 0;

  /*
   * init MSR data 
   */
  MSR1_STATUS = 0;
  pMSR1Data = 0;
  MSR2_STATUS = 0;
  pMSR2Data = 0;
  MSR3_STATUS = 0;
  pMSR3Data = 0;

  BuzzLen = 0xff;

  MSRLen = 0;
  MSRState = 0;
  MSR1Len = 0;
  MSR2Len = 0;
  MSR3Len = 0;

  GHmod = 0x00;
  GHcode = 0x00;
  GTmod = 0x00;
  GTcode = 0x00;

  H1mod = 0x00;
  H1code = 0x00;
  T1mod = 0x00;
  T1code = 0x00;

  H2mod = 0x00;
  H2code = 0x00;
  T2mod = 0x00;
  T2code = 0x00;

  H3mod = 0x00;
  H3code = 0x00;
  T3mod = 0x00;
  T3code = 0x00;

  // init timeout counter
  init_timer (&jpos_timer);
  jpos_timer.function = reinit_jpos;
  jpos_timer.data = 0;

  jpos_timer.expires = jiffies + HZ/50;	// 1 sec
  add_timer (&jpos_timer);
//  del_timer_sync (&jpos_timer);
//  mod_timer (&jpos_timer, jiffies + 1 * HZ);

  return 0;
}

/*
 * Cleanup - unregister the appropriate file from /proc 
 */
void
cleanup_msr ()
{
  del_timer_sync (&jpos_timer);
//  del_timer (&jpos_timer);

  if (Device_Major == 0)
    printk ("Error because Device_Major is 0\n");

  /*
   * Unregister the device 
   */
  unregister_chrdev (Device_Major, DEVICE_NAME);
}
