/*
 *      raw2kbd.c
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation; either version 2 of
 *	the License, or (at your option) any later version.
 *
 *
 * Decode the raw msrdata from buffer and forward it as keyboard inputs
 *
 */


#include <linux/kernel.h>	/* We're doing kernel work */
#include <linux/module.h>	/* Specifically, a module */

#include <linux/slab.h>
#include <linux/input.h>	// um Keys weiterzuleiten


#include "decode.h"
#include "scancodes.h"
#include "raw2kbd.h"


#include "header_terminator.h"


#define TR1START 3
#define TR2START 90
#define TR3START 139

//#define MY_DEBUG

typedef struct
{
  struct input_dev *dev;
  unsigned short keycode;
} bufferdata;

static int readpos = 0;
static int writepos = 0;
static bufferdata buffer[1024];



struct timer_list delyd_timer;

/* Timer routine for timeout function */
void
output_timedout (unsigned long ptr)
{
#ifdef MY_DEBUG
  printk("Key = %d Keypressed = %d readpos = %d\n",(buffer[readpos].keycode & 255),(buffer[readpos].keycode >> 8),readpos);
  printk("output_timedout readpos = %d\n",readpos);
#endif
  input_report_key (buffer[readpos].dev, (buffer[readpos].keycode & 255), (buffer[readpos].keycode >> 8));
  readpos++;
  if (readpos >= (sizeof (buffer) / sizeof (bufferdata)))
    {
      readpos = 0;
    }
  if (readpos == writepos)
    {				/* stop the timer */
#ifdef MY_DEBUG
      printk ("Delete Timer\n");
#endif
      del_timer (&delyd_timer);
    }
  else
    {				/* retrigger timer */
      mod_timer (&delyd_timer, jiffies + (EventDelay * HZ)/1000);
    }
}

void
output_init (void)
{
#ifdef MY_DEBUG
      printk ("output_init\n");
#endif
  init_timer (&delyd_timer);
  delyd_timer.function = output_timedout;
  delyd_timer.data = 0;
//  delyd_timer.data = (int)(&buffer[0]);
  delyd_timer.expires = jiffies + (EventDelay * HZ) / 1000;
  add_timer (&delyd_timer);
  del_timer_sync (&delyd_timer);
  mod_timer (&delyd_timer, jiffies + (EventDelay * HZ)/1000);	//retrigger timeout counter
}

void
intobuffer (struct input_dev *idev, int value)
{
#ifdef MY_DEBUG
      printk ("intobuffer value = %x writepos = %d\n",value,writepos);
#endif
  if (readpos == writepos)
    {				/* restart the timer */
      output_init ();
    }
  buffer[writepos].keycode = value;
  buffer[writepos].dev = idev;
  writepos++;
  if (writepos >= (sizeof (buffer) / sizeof (bufferdata)))
    {
      writepos = 0;
    }
}


void
inverttrack (char *orgtrack, char *reversetrack, unsigned char size)
{
  int i, r;
  for (i = 0; i < size; i++)
    {
      r = size - i - 1;
      reversetrack[r] = 0;
      if (orgtrack[i] & 0x01)
	reversetrack[r] |= 0x80;
      if (orgtrack[i] & 0x02)
	reversetrack[r] |= 0x40;
      if (orgtrack[i] & 0x04)
	reversetrack[r] |= 0x20;
      if (orgtrack[i] & 0x08)
	reversetrack[r] |= 0x10;
      if (orgtrack[i] & 0x10)
	reversetrack[r] |= 0x08;
      if (orgtrack[i] & 0x20)
	reversetrack[r] |= 0x04;
      if (orgtrack[i] & 0x40)
	reversetrack[r] |= 0x02;
      if (orgtrack[i] & 0x80)
	reversetrack[r] |= 0x01;
    }
}

void
presskeys (struct input_dev *idev, unsigned short kcode)
{
  if (kcode & 0x400)
    {
      intobuffer (idev,KEY_RIGHTALT + DOWN);
    }
  if (kcode & 0x200)
    {
      intobuffer (idev,KEY_LEFTCTRL + DOWN);
    }
  if (kcode & 0x100)
    {
      intobuffer (idev,KEY_LEFTSHIFT + DOWN);
    }
  intobuffer (idev,(kcode & 255) + DOWN);
  intobuffer (idev,(kcode & 255) + UP);

  if (kcode & 0x100)
    {
      intobuffer (idev,KEY_LEFTSHIFT + UP);
    }

  if (kcode & 0x200)
    {
      intobuffer (idev,KEY_LEFTCTRL + UP);
    }

  if (kcode & 0x400)
    {
      intobuffer (idev,KEY_RIGHTALT + UP);
    }
}

void
raw2kbd (struct input_dev *idev, unsigned char *msr_data)
{
  int result;
  int length;
  char pCARD_TYPE;
  unsigned char pbStartSentinel;
  unsigned char pbEndSentinel;
  unsigned char decoded[256];
  unsigned char reversetrack[256];
  int i;

  CDecode_CDecode ();

  if ((msr_data[0]) && (TRACK_1_ENABLE))
    {
      for (i = 0; i < (sizeof (Header_Track1) / sizeof (int)); i++)
	{
	  intobuffer (idev,Header_Track1[i]);
	}
#ifdef MY_DEBUG
      printk ("MSR 1 Raw Data =");
      for (i = 0; i < msr_data[0]; i++)
	{
	  printk (" %x", msr_data[i + TR1START]);
	}
      printk ("\n");
#endif
      inverttrack (&msr_data[TR1START], &reversetrack[0], msr_data[0]);
      result =
	CDecode_InterpreteSingleTrack (1, msr_data[0], &reversetrack[0],
				       &pbStartSentinel, &pbEndSentinel, 1,
				       &pCARD_TYPE);
      if (result != MSR_S_OK)
	result =
	  CDecode_InterpreteSingleTrack (1, msr_data[0], &msr_data[TR1START],
					 &pbStartSentinel, &pbEndSentinel, 1,
					 &pCARD_TYPE);
      length = CDecode_GetTrackData (decoded);
      if(START_STOP_SENTINALS) {
	  presskeys (idev,kbd_scancode[pbStartSentinel]);
      }
      for (i = 0; i < length; i++)
	{
	  presskeys (idev,kbd_scancode[decoded[i]]);
	}
      if(START_STOP_SENTINALS) {
	  presskeys (idev,kbd_scancode[pbEndSentinel]);
      }
      for (i = 0; i < (sizeof (Terminator_Track1) / sizeof (int)); i++)
	{
	  intobuffer (idev,Terminator_Track1[i]);
	}
    }

  if ((msr_data[1]) && (TRACK_2_ENABLE))
    {
      for (i = 0; i < (sizeof (Header_Track2) / sizeof (int)); i++)
	{
	  intobuffer (idev,Header_Track2[i]);
	}
#ifdef MY_DEBUG
      printk ("MSR 2 Raw Data =");
      for (i = 0; i < msr_data[1]; i++)
	{
	  printk (" %x", msr_data[i + TR2START]);
	}
      printk ("\n");
#endif
      inverttrack (&msr_data[TR2START], &reversetrack[0], msr_data[1]);
      result =
	CDecode_InterpreteSingleTrack (2, msr_data[1], &reversetrack[0],
				       &pbStartSentinel, &pbEndSentinel, 1,
				       &pCARD_TYPE);
      if (result != MSR_S_OK)
	result =
	  CDecode_InterpreteSingleTrack (2, msr_data[1], &msr_data[TR2START],
					 &pbStartSentinel, &pbEndSentinel, 1,
					 &pCARD_TYPE);
      length = CDecode_GetTrackData (decoded);
      if(START_STOP_SENTINALS) {
	  presskeys (idev,kbd_scancode[pbStartSentinel]);
      }
      for (i = 0; i < length; i++)
	{
	  presskeys (idev,kbd_scancode[decoded[i]]);
	}
      if(START_STOP_SENTINALS) {
	  presskeys (idev,kbd_scancode[pbEndSentinel]);
      }
      for (i = 0; i < (sizeof (Terminator_Track2) / sizeof (int)); i++)
	{
	  intobuffer (idev,Terminator_Track2[i]);
	}
    }

  if ((msr_data[2]) && (TRACK_3_ENABLE))
    {
      for (i = 0; i < (sizeof (Header_Track3) / sizeof (int)); i++)
	{
	  intobuffer (idev,Header_Track3[i]);
	}
#ifdef MY_DEBUG
      printk ("MSR 3 Raw Data =");
      for (i = 0; i < msr_data[2]; i++)
	{
	  printk (" %x", msr_data[i + TR3START]);
	}
      printk ("\n");
#endif
      inverttrack (&msr_data[TR3START], &reversetrack[0], msr_data[2]);
      result =
	CDecode_InterpreteSingleTrack (3, msr_data[2], &reversetrack[0],
				       &pbStartSentinel, &pbEndSentinel, 1,
				       &pCARD_TYPE);
      if (result != MSR_S_OK)
	result =
	  CDecode_InterpreteSingleTrack (3, msr_data[2], &msr_data[TR3START],
					 &pbStartSentinel, &pbEndSentinel, 1,
					 &pCARD_TYPE);
      length = CDecode_GetTrackData (decoded);
      if(START_STOP_SENTINALS) {
	  presskeys (idev,kbd_scancode[pbStartSentinel]);
      }
      for (i = 0; i < length; i++)
	{
	  presskeys (idev,kbd_scancode[decoded[i]]);
	}
      if(START_STOP_SENTINALS) {
	  presskeys (idev,kbd_scancode[pbEndSentinel]);
      }
      for (i = 0; i < (sizeof (Terminator_Track3) / sizeof (int)); i++)
	{
	  intobuffer (idev,Terminator_Track3[i]);
	}
    }
}
