/*
 * $Id: usbmsr.c,v 1.27 2001/12/27 10:37:41 vojtech Exp $
 *
 *  Copyright (c) 1999-2001 Vojtech Pavlik
 *
 *  USB HIDBP Keyboard support
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Should you need to contact me, the author, you can do so either by
 * e-mail - mail your message to <vojtech@ucw.cz>, or by paper mail:
 * Vojtech Pavlik, Simunkova 1594, Prague 8, 182 00 Czech Republic
 */

//#define MY_DEBUG

#define SYSSOUND

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/usb.h>
#include <linux/version.h>

#include "msr_devs.h"

#define USB_MSR_VENDOR_ID	0x046a
#define USB_MSR_PRODUCT_ID	0x000f

/*
 * Version Information
 */
#define DRIVER_VERSION ""
#define DRIVER_AUTHOR "Vojtech Pavlik <vojtech@ucw.cz>"
#define DRIVER_DESC "USB HID MSR driver"
#define DRIVER_LICENSE "GPL"

MODULE_AUTHOR (DRIVER_AUTHOR);
MODULE_DESCRIPTION (DRIVER_DESC);
MODULE_LICENSE (DRIVER_LICENSE);

struct usb_msr
{
  struct input_dev * dev;
  struct usb_device *usbdev;
  unsigned char old[9];
  struct urb *irq, *beep, *klpos, *oldbeep;
  unsigned short newbeeps;
  char name[128];
  char phys[64];
  int open;

  char msrdata[257];
  unsigned char urbcount;

  unsigned char *new;
  struct usb_ctrlrequest *cr, *kl, *ob;
  unsigned short *beeps, *getklpos;
  unsigned char *oldbeeps;
  dma_addr_t cr_dma;
  dma_addr_t kl_dma;
  dma_addr_t ob_dma;
  dma_addr_t new_dma;
  dma_addr_t beeps_dma;
  dma_addr_t oldbeeps_dma;
  dma_addr_t getklpos_dma;
  unsigned char report_id1;
  unsigned char report_id2;
  struct timer_list jiq_timer;
  unsigned char oldkeyboard;
  unsigned char apusb;
};





static void
usb_msr_irq (struct urb *urb)
{
  unsigned char beeptime;

  struct usb_msr *msr = urb->context;
  int i;

#ifdef MY_DEBUG
  printk ("usb_msr_irq\n");
#endif
  if (msr->apusb)
    {

      switch (urb->status)
	{
	case 0:		/* success */
	  break;
	case -ECONNRESET:	/* unlink */
	case -ENOENT:
	case -ESHUTDOWN:
	  msr->urbcount = 0;
	  return;
	  /* -EPIPE:  should clear the halt */
	default:		/* error */
	  msr->urbcount = 0;
	  goto resubmit;
	}

#ifdef MY_DEBUG
      mod_timer (&msr->jiq_timer, jiffies + 2 * HZ);	//retrigger timeout counter 2secs
#else
      mod_timer (&msr->jiq_timer, jiffies + HZ / 10);	//retrigger timeout counter 0.1secs
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
    input_sync (msr->dev);
#else
#endif


      if (msr->urbcount == 0)
	{
	  msr->report_id1 = 0;
	  msr->report_id2 = 0;

	  if (msr->new[0] == 1)
	    {
	      msr->report_id1 = 1;
	    }
	  if (msr->new[0] == 2)
	    {
	      msr->report_id2 = 1;
	    }
	}

      if (msr->urbcount == 1)
	{
#ifdef MY_DEBUG
	  printk ("switch off beeper\n");
#endif
	  input_event (msr->dev, EV_SND, SND_MAX, 0);
	}

#ifdef MY_DEBUG
      printk ("urb block %x received\n", msr->urbcount);
#endif

      if (msr->report_id2)
	{
	  msr->urbcount = 0;
	  keylock (msr->dev, msr->new[4]);
#ifdef MY_DEBUG
	  printk ("keylock position %x received\n", msr->new[4]);
	  memset (&msr->new[0], 0, 64);
#endif
	}
      else
	{

	  if (msr->urbcount < 4)
	    {
	      memcpy (&msr->msrdata[msr->urbcount * 64], &msr->new[0], 64);

	      memset (&msr->new[0], 0, 64);

	      msr->urbcount++;
	    }

	  if (msr->urbcount >= 4)
	    {
	      msr->urbcount = 0;

	      if (!msr->report_id1)
		{
		  beeptime = msrcheck (msr->dev, &msr->msrdata[0], 1);
		}
	      if (msr->report_id1)
		{
		  msr->msrdata[256] = 0;
		  beeptime = msrcheck (msr->dev, &msr->msrdata[1], 1);
		}
	      if (beeptime)
		{
		  input_event (msr->dev, EV_SND, SND_MAX, beeptime);
		}

	    }
	}


	input_sync (msr->dev);


    }
  else
    {
      switch (urb->status)
	{
	case 0:		/* success */
	  break;
	case -ECONNRESET:	/* unlink */
	case -ENOENT:
	case -ESHUTDOWN:
	  return;
	  /* -EPIPE:  should clear the halt */
	default:		/* error */
	  goto resubmit;
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
      input_regs (msr->dev, regs);
#else
#endif

//      for (i = 3; i < 9; i++)
//	{
//	  if (msr->new[i] > 3
//	      && memscan (msr->old + 3, msr->new[i], 6) == msr->old + 9)
//	    {
//	      msr->msrdata[0] = msr->new[1];
//	      msr->msrdata[1] = msr->new[i];
//	      msrcheck (&msr->dev, &msr->msrdata[0], 0);
//	    }
//	}
//
//      input_sync (&msr->dev);
//
//      memcpy (msr->old, msr->new, 9);


      memcpy (&msr->msrdata[0], msr->new, 9);
      input_sync (msr->dev);
      msrcheck (msr->dev, &msr->msrdata[0], 0);


    }
resubmit:
  i = usb_submit_urb (urb, GFP_ATOMIC);
  if (i)
    err ("can't resubmit intr, %s-%s/input0, status %d",
	 msr->usbdev->bus->bus_name, msr->usbdev->devpath, i);


}

int
usb_msr_event (struct input_dev *dev, unsigned int type, unsigned int code,
	       int value)
{
  struct usb_msr *msr = input_get_drvdata(dev);

#ifdef MY_DEBUG
  printk ("usb_msr_event\n");
#endif

  if (msr->apusb)
    {

      if (type != EV_SND)
	return -1;


      if (code == SND_TONE)
	return -1;
      if (code == SND_CLICK)
	return -1;

#ifdef SYSSOUND
      if ((code == SND_BELL) && value)
	value = 0xff;
#else
      if (code == SND_BELL)
	return -1;
#endif

      if (msr->oldkeyboard)
	{
	  if (msr->oldbeep->status == -EINPROGRESS)
	    return 0;

	  if (value)
	    {
	      msr->newbeeps = value;
	    }
	  else
	    {
	      msr->newbeeps = 0x01;
	    }


	  if (*(msr->oldbeeps) == msr->newbeeps)
	    return 0;

	  *(msr->oldbeeps) = msr->newbeeps;
	  msr->oldbeep->dev = msr->usbdev;
	  if (usb_submit_urb (msr->oldbeep, GFP_ATOMIC))
	    err ("usb_submit_urb(beeps) failed");
	}
      else
	{
	  if (msr->beep->status == -EINPROGRESS)
	    return 0;

	  if (value)
	    {
	      msr->newbeeps = 0x01 | (value << 8);
	    }
	  else
	    {
	      msr->newbeeps = 0x0101;
	    }


	  if (*(msr->beeps) == msr->newbeeps)
	    return 0;

	  *(msr->beeps) = msr->newbeeps;
	  msr->beep->dev = msr->usbdev;
	  if (usb_submit_urb (msr->beep, GFP_ATOMIC))
	    err ("usb_submit_urb(beeps) failed");
	}

    }
  else
    {
      return -1;

    }

  return 0;
}

static void
usb_msr_beep (struct urb *urb)
{
  struct usb_msr *msr = urb->context;

#ifdef MY_DEBUG
  printk ("usb_msr_beep\n");
#endif

  if (urb->status)
    warn ("beep urb status %d received", urb->status);

  if (*(msr->beeps) == msr->newbeeps)
    return;

  *(msr->beeps) = msr->newbeeps;
  msr->beep->dev = msr->usbdev;
  if (usb_submit_urb (msr->beep, GFP_ATOMIC))
    err ("usb_submit_urb(beeps) failed");
}

static void
usb_msr_oldbeep (struct urb *urb)
{
  struct usb_msr *msr = urb->context;

#ifdef MY_DEBUG
  printk ("usb_msr_oldbeep\n");
#endif

  if (urb->status)
    warn ("beep urb status %d received", urb->status);

  if (*(msr->oldbeeps) == msr->newbeeps)
    return;

  *(msr->oldbeeps) = msr->newbeeps;
  msr->oldbeep->dev = msr->usbdev;
  if (usb_submit_urb (msr->oldbeep, GFP_ATOMIC))
    err ("usb_submit_urb(beeps) failed");
}


static void
usb_msr_klpos (struct urb *urb)
{
  struct usb_msr *msr = urb->context;

#ifdef MY_DEBUG
  printk ("usb_msr_klpos\n");
#endif

  if (urb->status)
    {
      warn ("klpos urb status %d received", urb->status);
      msr->oldkeyboard = 1;
    }
  else
    {
      msr->oldkeyboard = 0;
    }

  if (*(msr->getklpos) == 0x0102)
    return;

  *(msr->getklpos) = 0x0102;
  msr->klpos->dev = msr->usbdev;
  if (usb_submit_urb (msr->klpos, GFP_ATOMIC))
    err ("usb_submit_urb(klpos) failed");
}

/* Timer routine for timeout function */
void
jiq_timedout (unsigned long ptr)
{
  struct usb_msr *msr;
  msr = (struct usb_msr *) ptr;
  // add commands for timeout handling
#ifdef MY_DEBUG
  printk ("Timeout\n");
#endif
  msr->urbcount = 0;

}


static int
usb_msr_open (struct input_dev *dev)
{
  struct usb_msr *msr = input_get_drvdata(dev);

#ifdef MY_DEBUG
  printk ("usb_msr_open\n");
#endif

  if (msr->open++)
    return 0;

  msr->irq->dev = msr->usbdev;
  if (usb_submit_urb (msr->irq, GFP_KERNEL))
    {
      msr->open--;
      return -EIO;
    }
  if (msr->apusb)
    {
      if (usb_submit_urb (msr->klpos, GFP_KERNEL))
	{
	  msr->open--;
	  return -EIO;
	}

      // init timeout counter
      init_timer (&msr->jiq_timer);
      msr->jiq_timer.function = jiq_timedout;
      msr->jiq_timer.data = (unsigned long) msr;

      msr->jiq_timer.expires = jiffies + 1 * HZ;	// 1 sec
      add_timer (&msr->jiq_timer);
      del_timer_sync (&msr->jiq_timer);
    }
  return 0;
}

static void
usb_msr_close (struct input_dev *dev)
{
  struct usb_msr *msr = input_get_drvdata(dev);

#ifdef MY_DEBUG
  printk ("usb_msr_close\n");
#endif
  if (msr->apusb)
    {
      // delete timeout counter
      del_timer_sync (&msr->jiq_timer);
    }
  if (!--msr->open)
    usb_unlink_urb (msr->irq);
}

static int
usb_msr_alloc_mem (struct usb_device *dev, struct usb_msr *msr)
{
	msr->dev = NULL;
	// Allocate mem->dev
	#ifdef MY_DEBUG
	printk("input_allocate_device()\n");
	#endif
	msr->dev = input_allocate_device();   
	if ( NULL == msr->dev)
	{
		printk("input_allocate_device_failed!\n");
		return -1;
	}

#ifdef MY_DEBUG
  printk ("usb_msr_alloc_mem\n");
#endif
  if (!(msr->irq = usb_alloc_urb (0, GFP_KERNEL)))
    return -1;
#ifdef MY_DEBUG
  printk ("irq urb allocated\n");
#endif
//if (msr->apusb){
  if (!(msr->beep = usb_alloc_urb (0, GFP_KERNEL)))
    return -1;
#ifdef MY_DEBUG
  printk ("beep report id 1 urb allocated\n");
#endif
  if (!(msr->klpos = usb_alloc_urb (0, GFP_KERNEL)))
    return -1;
#ifdef MY_DEBUG
  printk ("keylock position urb allocated\n");
#endif
  if (!(msr->oldbeep = usb_alloc_urb (0, GFP_KERNEL)))
    return -1;
#ifdef MY_DEBUG
  printk ("beep report id 0 urb allocated\n");
#endif
//}
  if (!(msr->new = usb_buffer_alloc (dev, 64, GFP_ATOMIC, &msr->new_dma)))
    return -1;
#ifdef MY_DEBUG
  printk ("new buffer allocated\n");
#endif
  if (!
      (msr->cr =
       usb_buffer_alloc (dev, sizeof (struct usb_ctrlrequest), GFP_ATOMIC,
			 &msr->cr_dma)))
    return -1;
#ifdef MY_DEBUG
  printk ("cr buffer allocated\n");
#endif
//if (msr->apusb){
  if (!
      (msr->kl =
       usb_buffer_alloc (dev, sizeof (struct usb_ctrlrequest), GFP_ATOMIC,
			 &msr->kl_dma)))
    return -1;
#ifdef MY_DEBUG
  printk ("cr buffer allocated\n");
#endif
  if (!
      (msr->ob =
       usb_buffer_alloc (dev, sizeof (struct usb_ctrlrequest), GFP_ATOMIC,
			 &msr->ob_dma)))
    return -1;
#ifdef MY_DEBUG
  printk ("beep buffer allocated\n");
#endif
  if (!(msr->beeps = usb_buffer_alloc (dev, 1, GFP_ATOMIC, &msr->beeps_dma)))
    return -1;
#ifdef MY_DEBUG
  printk ("beeps buffer allocated\n");
#endif
  if (!
      (msr->getklpos =
       usb_buffer_alloc (dev, 1, GFP_ATOMIC, &msr->getklpos_dma)))
    return -1;
#ifdef MY_DEBUG
  printk ("getklpos buffer allocated\n");
#endif
  if (!
      (msr->oldbeeps =
       usb_buffer_alloc (dev, 1, GFP_ATOMIC, &msr->oldbeeps_dma)))
    return -1;
#ifdef MY_DEBUG
  printk ("oldbeeps buffer allocated\n");
#endif
//}
  return 0;
}

static void
usb_msr_free_mem (struct usb_device *dev, struct usb_msr *msr)
{

#ifdef MY_DEBUG
  printk ("usb_msr_free_mem\n");
#endif

  if (msr->dev)
	{
		input_free_device(msr->dev);
		msr->dev = 0;
	}

  if (msr->irq)
    usb_free_urb (msr->irq);
//if (msr->apusb){
  if (msr->beep)
    usb_free_urb (msr->beep);
  if (msr->klpos)
    usb_free_urb (msr->klpos);
  if (msr->oldbeep)
    usb_free_urb (msr->oldbeep);
//}
  if (msr->new)
    usb_buffer_free (dev, 64, msr->new, msr->new_dma);
  if (msr->cr)
    usb_buffer_free (dev, sizeof (struct usb_ctrlrequest), msr->cr,
		     msr->cr_dma);
//if (msr->apusb){
  if (msr->kl)
    usb_buffer_free (dev, sizeof (struct usb_ctrlrequest), msr->kl,
		     msr->kl_dma);
  if (msr->ob)
    usb_buffer_free (dev, sizeof (struct usb_ctrlrequest), msr->ob,
		     msr->ob_dma);
  if (msr->beeps)
    usb_buffer_free (dev, 1, msr->beeps, msr->beeps_dma);
  if (msr->getklpos)
    usb_buffer_free (dev, 1, msr->getklpos, msr->getklpos_dma);
  if (msr->oldbeeps)
    usb_buffer_free (dev, 1, msr->oldbeeps, msr->oldbeeps_dma);
//}
}

static int
usb_msr_probe (struct usb_interface *iface, const struct usb_device_id *id)
{
  struct usb_device *dev = interface_to_usbdev (iface);
  struct usb_host_interface *interface;
  struct usb_endpoint_descriptor *endpoint;
  struct usb_msr *msr;
  int i, pipe, maxp;
  char path[64];
  char *buf;

#ifdef MY_DEBUG
  printk ("usb_msr_probe\n");
#endif

  interface = iface->cur_altsetting;

  if (dev->descriptor.idVendor != USB_MSR_VENDOR_ID)
    {
      return -ENODEV;
    }

  if (!(dev->descriptor.idProduct == 0x0f ||
	dev->descriptor.idProduct == 0x16 ||
	dev->descriptor.idProduct == 0x17 ||
	dev->descriptor.idProduct == 0x18 ||
	dev->descriptor.idProduct == 0x19 ||
	dev->descriptor.idProduct == 0x1a ||
	dev->descriptor.idProduct == 0x1b ||
	dev->descriptor.idProduct == 0x1c ||
	dev->descriptor.idProduct == 0x1d ||
	dev->descriptor.idProduct == 0x1e ||
	dev->descriptor.idProduct == 0x1f ||
	dev->descriptor.idProduct == 0x34 ||
	dev->descriptor.idProduct == 0x35 ||
	dev->descriptor.idProduct == 0x36 ||
	dev->descriptor.idProduct == 0x37 ||
	dev->descriptor.idProduct == 0x38 ||
	dev->descriptor.idProduct == 0x39 ||
	dev->descriptor.idProduct == 0x3a ||
	dev->descriptor.idProduct == 0x3b ||
	dev->descriptor.idProduct == 0x40 ||
	dev->descriptor.idProduct == 0x41 ||
	dev->descriptor.idProduct == 0x42 ||
	dev->descriptor.idProduct == 0x43))
    {

      return -ENODEV;
    }





  if (interface->desc.bNumEndpoints != 1)
    return -ENODEV;

  endpoint = &interface->endpoint[0].desc;
  if (!(endpoint->bEndpointAddress & 0x80))
    return -ENODEV;
  if ((endpoint->bmAttributes & 3) != 3)
    return -ENODEV;

  pipe = usb_rcvintpipe (dev, endpoint->bEndpointAddress);
  maxp = usb_maxpacket (dev, pipe, usb_pipeout (pipe));

  if (!(msr = kmalloc (sizeof (struct usb_msr), GFP_KERNEL)))
    return -ENOMEM;
  memset (msr, 0, sizeof (struct usb_msr));

  if (usb_msr_alloc_mem (dev, msr))
    {
#ifdef MY_DEBUG
      printk ("usb_msr_alloc_mem failed\n");
#endif
      usb_msr_free_mem (dev, msr);
      kfree (msr);
      return -ENOMEM;
    }

  if ((dev->descriptor.idProduct == 0x0f ||
       dev->descriptor.idProduct == 0x16 ||
       dev->descriptor.idProduct == 0x17 ||
       dev->descriptor.idProduct == 0x18 ||
       dev->descriptor.idProduct == 0x19 ||
       dev->descriptor.idProduct == 0x1a))
    {
      msr->apusb = 1;
    }

  msr->usbdev = dev;

//      msr->dev.evbit[0] = BIT(EV_KEY) | BIT(EV_SND) | BIT(EV_REP);
//      msr->dev.sndbit[0] = BIT(SND_BELL) | BIT(SND_CLICK) | BIT(SND_TONE);

	#ifdef MY_DEBUG
		printk("Set evBits");
	#endif


  if (msr->apusb)
    {
      msr->dev->evbit[0] = BIT (EV_KEY) | BIT (EV_SND);

#ifdef SYSSOUND
      msr->dev->sndbit[0] = BIT (SND_BELL) | BIT (SND_MAX);
#else
      msr->dev->sndbit[0] = BIT (SND_MAX);
#endif
    }
  else
    {
      msr->dev->evbit[0] = BIT (EV_KEY);
    }

  for (i = 0; i < 255; i++)
    {
      set_bit (i, msr->dev->keybit);
//              set_bit(usb_msr_keycode[i], msr->dev.keybit);
    }
  clear_bit (0, msr->dev->keybit);

  input_set_drvdata(msr->dev, msr);
  msr->dev->event = usb_msr_event;
  msr->dev->open = usb_msr_open;
  msr->dev->close = usb_msr_close;
	
	#ifdef MY_DEBUG
		printk("Set APUSB Options\n");
	#endif


  if (msr->apusb)
    {
      usb_fill_int_urb (msr->irq, dev, pipe,
			msr->new, (maxp > 64 ? 64 : maxp),
			usb_msr_irq, msr, endpoint->bInterval);
      msr->irq->transfer_dma = msr->new_dma;
      msr->irq->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

      msr->cr->bRequestType = USB_TYPE_CLASS | USB_RECIP_INTERFACE;
      msr->cr->bRequest = 0x09;
      msr->cr->wValue = cpu_to_le16 (0x201);
      msr->cr->wIndex = cpu_to_le16 (interface->desc.bInterfaceNumber);
      msr->cr->wLength = cpu_to_le16 (2);

      msr->kl->bRequestType = USB_TYPE_CLASS | USB_RECIP_INTERFACE;
      msr->kl->bRequest = 0x09;
      msr->kl->wValue = cpu_to_le16 (0x202);
      msr->kl->wIndex = cpu_to_le16 (interface->desc.bInterfaceNumber);
      msr->kl->wLength = cpu_to_le16 (2);

      msr->ob->bRequestType = USB_TYPE_CLASS | USB_RECIP_INTERFACE;
      msr->ob->bRequest = 0x09;
      msr->ob->wValue = cpu_to_le16 (0x200);
      msr->ob->wIndex = cpu_to_le16 (interface->desc.bInterfaceNumber);
      msr->ob->wLength = cpu_to_le16 (1);

	#ifdef MY_DEBUG
		printk("usb_make_path\n");
	#endif

      usb_make_path (dev, path, 64);
      sprintf (msr->phys, "%s/input0", path);
	#ifdef MY_DEBUG
		printk("usb_make_path end\n");
	#endif

      msr->dev->name = msr->name;
      msr->dev->phys = msr->phys;
      msr->dev->id.bustype = BUS_USB;
      msr->dev->id.vendor = dev->descriptor.idVendor;
      msr->dev->id.product = dev->descriptor.idProduct;
      msr->dev->id.version = dev->descriptor.bcdDevice;


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
    msr->dev->dev = &iface->dev;
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26))
    msr->dev->cdev.dev = &iface->dev;
#endif

	#ifdef MY_DEBUG
		printk("set msr->dev.cdev \n");
	#endif

  if (!(buf = kmalloc (63, GFP_KERNEL)))
	{
	  usb_free_urb (msr->irq);
		usb_msr_free_mem (dev, msr);

	  kfree (msr);
	  return -ENOMEM;
	}

  if (dev->descriptor.iManufacturer &&
	  usb_string (dev, dev->descriptor.iManufacturer, buf, 63) > 0)
	strcat (msr->name, buf);
      if (dev->descriptor.iProduct &&
	  usb_string (dev, dev->descriptor.iProduct, buf, 63) > 0)
	sprintf (msr->name, "%s %s", msr->name, buf);

      if (!strlen (msr->name))
	sprintf (msr->name, "USB HID Magetic Stripe Reader %04x:%04x",
		 msr->dev->id.vendor, msr->dev->id.product);

      kfree (buf);


	#ifdef MY_DEBUG
		printk("usb_fill_control_urb \n");
	#endif

      usb_fill_control_urb (msr->beep, dev, usb_sndctrlpipe (dev, 0),
			    (void *) msr->cr, msr->beeps, 2,
			    usb_msr_beep, msr);
      msr->beep->setup_dma = msr->cr_dma;
      msr->beep->transfer_dma = msr->beeps_dma;
      msr->beep->transfer_flags |= (URB_NO_TRANSFER_DMA_MAP
				    | URB_NO_SETUP_DMA_MAP);

	#ifdef MY_DEBUG
		printk("usb_fill_control_urb \n");
	#endif

      usb_fill_control_urb (msr->oldbeep, dev, usb_sndctrlpipe (dev, 0),
			    (void *) msr->ob, msr->oldbeeps, 1,
			    usb_msr_oldbeep, msr);
      msr->oldbeep->setup_dma = msr->ob_dma;
      msr->oldbeep->transfer_dma = msr->oldbeeps_dma;
      msr->oldbeep->transfer_flags |= (URB_NO_TRANSFER_DMA_MAP
				       | URB_NO_SETUP_DMA_MAP);

	#ifdef MY_DEBUG
		printk("usb_fill_control_urb \n");
	#endif
		
      usb_fill_control_urb (msr->klpos, dev, usb_sndctrlpipe (dev, 0),
			    (void *) msr->kl, msr->getklpos, 2,
			    usb_msr_klpos, msr);

      msr->klpos->setup_dma = msr->kl_dma;
      msr->klpos->transfer_dma = msr->getklpos_dma;
      msr->klpos->transfer_flags |= (URB_NO_TRANSFER_DMA_MAP
				     | URB_NO_SETUP_DMA_MAP);


    }
  else
    {
      usb_fill_int_urb (msr->irq, dev, pipe,
			msr->new, (maxp > 9 ? 9 : maxp),
			usb_msr_irq, msr, endpoint->bInterval);
      msr->irq->transfer_dma = msr->new_dma;
      msr->irq->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

      msr->cr->bRequestType = USB_TYPE_CLASS | USB_RECIP_INTERFACE;
      msr->cr->bRequest = 0x09;
      msr->cr->wValue = cpu_to_le16 (0x200);
      msr->cr->wIndex = cpu_to_le16 (interface->desc.bInterfaceNumber);
      msr->cr->wLength = cpu_to_le16 (1);

      usb_make_path (dev, path, 64);
      sprintf (msr->phys, "%s/input0", path);

      msr->dev->name = msr->name;
      msr->dev->phys = msr->phys;
      msr->dev->id.bustype = BUS_USB;
      msr->dev->id.vendor = dev->descriptor.idVendor;
      msr->dev->id.product = dev->descriptor.idProduct;
      msr->dev->id.version = dev->descriptor.bcdDevice;


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
    msr->dev->dev = &iface->dev;
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26))
    msr->dev->cdev.dev = &iface->dev;
#endif
 


  if (!(buf = kmalloc (63, GFP_KERNEL)))
	{
	  usb_free_urb (msr->irq);
	  usb_msr_free_mem (dev, msr);

	  kfree (msr);
	  return -ENOMEM;
	}

	     if (dev->descriptor.iManufacturer &&
	  usb_string (dev, dev->descriptor.iManufacturer, buf, 63) > 0)
	strcat (msr->name, buf);
      if (dev->descriptor.iProduct &&
	  usb_string (dev, dev->descriptor.iProduct, buf, 63) > 0)
	sprintf (msr->name, "%s %s", msr->name, buf);

      if (!strlen (msr->name))
	sprintf (msr->name, "USB HIDBP Keyboard %04x:%04x",
		 msr->dev->id.vendor, msr->dev->id.product);

      kfree (buf);


    }

	#ifdef MY_DEBUG
		printk(KERN_INFO "next call is input_register_device. Addr= %x \n", (unsigned int)msr->dev );
	#endif	

  if (input_register_device (msr->dev))
  {
    printk(KERN_INFO "Unable to register input device\n");
  }

  printk (KERN_INFO "input: %s on %s\n", msr->name, path);

  usb_set_intfdata (iface, msr);
  return 0;
}

static void
usb_msr_disconnect (struct usb_interface *intf)
{
  struct usb_msr *msr = usb_get_intfdata (intf);

#ifdef MY_DEBUG
  printk ("usb_msr_disconnect\n");
#endif

  usb_set_intfdata (intf, NULL);
  if (msr)
    {
      usb_unlink_urb (msr->irq);
      input_unregister_device (msr->dev);
			/*do not free memory here!*/
			msr->dev = NULL;
			usb_msr_free_mem (interface_to_usbdev (intf), msr);
      
      kfree (msr);
    }
}

static struct usb_device_id usb_msr_id_table[] = {
  {USB_INTERFACE_INFO (3, 0, 0)},
  {}				/* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, usb_msr_id_table);

static struct usb_driver usb_msr_driver = {
  .name = "chyposusbmsr",
  .probe = usb_msr_probe,
  .disconnect = usb_msr_disconnect,
  .id_table = usb_msr_id_table,
};

static int __init
usb_msr_init (void)
{
#ifdef MY_DEBUG
  printk ("pre init_msr\n");
#endif

  init_msr ();

#ifdef MY_DEBUG
  printk ("usb_msr_init\n");
#endif
  int result = usb_register (&usb_msr_driver);
  if (result == 0)
    info (DRIVER_VERSION ":" DRIVER_DESC);

  return result;
}

static void __exit
usb_msr_exit (void)
{

#ifdef MY_DEBUG
  printk ("usb_msr_exit\n");
#endif
  usb_deregister (&usb_msr_driver);
  cleanup_msr ();
}

module_init (usb_msr_init);
module_exit (usb_msr_exit);
